from abc import ABC, abstractmethod
from typing import Optional, List
from analogai.foundation.modules.notion.notion import Notion
from analogai.foundation.common.dtos.notion_expression import NotionExpression


class NotionRepository(ABC):
    @abstractmethod
    def create(self, notion: Notion) -> Notion:
        pass

    @abstractmethod
    def find(self, agent_id: str, expression: NotionExpression, context_ids: Optional[List[str]] = None) -> Optional[Notion]:
        pass

    @abstractmethod
    def find_by_id(self, notion_id: str) -> Optional[Notion]:
        pass

    @abstractmethod
    def find_all_by_agent(self, agent_id: str) -> List[Notion]:
        pass


import unittest
from unittest import TestCase
from unittest.mock import Mock
from analogai.foundation.modules.belief.belief import Belief
from analogai.foundation.modules.notion.notion import Notion
from analogai.foundation.modules.direct_statement.direct_statement import DirectStatement as Statement
from analogai.foundation.core.value_objects.relation import Relation

from analogai.foundation.modules.user.user import User
from analogai.foundation.modules.agent.agent import Agent
from analogai.foundation.infrastructure.repositories.in_memory.in_memory_belief_repository_impl import InMemoryBeliefRepositoryImpl
from analogai.foundation.infrastructure.storage.storage_adapter import StorageAdapter
from analogai.foundation.infrastructure.mappers import BeliefStorageMapper, StatementStorageMapper
from analogai.foundation.settings import config

class TestInMemoryBeliefRepositoryImpl(TestCase):
    def setUp(self):
        self._config_override = config.override(**{"vector.enabled": False})
        self._config_override.__enter__()
        self.storage_adapter = Mock(spec=StorageAdapter)
        self.repository = InMemoryBeliefRepositoryImpl(storage_adapter=self.storage_adapter)
        self.user = User(id="user_id", name="Test User", email="test@example.com")
        self.agent = Agent(id="agent_id", knowledge_base="Test KB", creator=self.user, roles=[])
        self.subject_notion = Notion(user_id=self.user.id, agent_id=self.agent.id, caption="human", id="human_id")
        self.complement_notion = Notion(user_id=self.user.id, agent_id=self.agent.id, caption="mortal", id="mortal_id")
        self.statement = Statement(
            user_id=self.user.id,
            agent_id=self.agent.id,
            subject_notion=self.subject_notion,
            relation=Relation.from_string("is"),
            complement_notion=self.complement_notion,
            probability=0.8,
            validity=0.9,
            id="stmt_id"
        )
        self.belief = Belief(agent_id=self.agent.id, id="belief_id")
        self.belief.add_statement(self.statement)

    def test_save_calls_store(self):
        saved = self.repository.save(self.belief)
        
        self.storage_adapter.store.assert_called_once()
        args, _ = self.storage_adapter.store.call_args
        self.assertEqual(args[0], "beliefs")
        self.assertEqual(saved.subject_notion.id, self.subject_notion.id)
        self.assertEqual(saved.complement_notion.id, self.complement_notion.id)

    def test_find(self):
        belief_data = BeliefStorageMapper.to_dict(self.belief)
        statement_data = StatementStorageMapper.to_dict(self.statement)
        self.storage_adapter.retrieve_by_attributes.return_value = [belief_data]

        def retrieve_by_id_side_effect(collection, id):
            if collection == "statements" and id == self.statement.id:
                return statement_data
            return None

        self.storage_adapter.retrieve_by_id.side_effect = retrieve_by_id_side_effect

        result = self.repository.find(self.subject_notion, self.complement_notion, Relation.from_string("is"))
        
        expected_query = {
            'subject_notion.id': "human_id",
            'complement_notion.id': "mortal_id",
            'relation': "is"
        }
        self.storage_adapter.retrieve_by_attributes.assert_called_once_with("beliefs", expected_query)
        self.assertEqual(result.id, self.belief.id)

    def test_find_no_result(self):
        self.storage_adapter.retrieve_by_attributes.return_value = []
        
        result = self.repository.find(self.subject_notion, self.complement_notion, Relation.from_string("is"))
        
        expected_query = {
            'subject_notion.id': "human_id",
            'complement_notion.id': "mortal_id",
            'relation': "is"
        }
        self.storage_adapter.retrieve_by_attributes.assert_called_once_with("beliefs", expected_query)
        self.assertIsNone(result)

    def test_find_by_id(self):
        belief_data = BeliefStorageMapper.to_dict(self.belief)
        statement_data = StatementStorageMapper.to_dict(self.statement)
        
        def retrieve_by_id_side_effect(collection, id):
            if collection == "beliefs":
                return belief_data
            elif collection == "statements":
                return statement_data
            return None
        
        self.storage_adapter.retrieve_by_id.side_effect = retrieve_by_id_side_effect
        
        result = self.repository.find_by_id("belief_id")
        
        self.assertEqual(self.storage_adapter.retrieve_by_id.call_count, 2)
        self.storage_adapter.retrieve_by_id.assert_any_call("beliefs", "belief_id")
        self.storage_adapter.retrieve_by_id.assert_any_call("statements", "stmt_id")
        self.assertEqual(result.id, self.belief.id)

    def test_find_by_id_not_found(self):
        self.storage_adapter.retrieve_by_id.return_value = None
        
        result = self.repository.find_by_id("nonexistent_id")
        
        self.storage_adapter.retrieve_by_id.assert_called_once_with("beliefs", "nonexistent_id")
        self.assertIsNone(result)

    def tearDown(self):
        if self._config_override:
            self._config_override.__exit__(None, None, None)

    def test_delete(self):
        self.repository.delete(self.belief)
        
        self.storage_adapter.delete.assert_called_once_with("beliefs", "belief_id")

    def test_search_by_notions(self):
        belief_data = BeliefStorageMapper.to_dict(self.belief)
        statement_data = StatementStorageMapper.to_dict(self.statement)
        self.storage_adapter.retrieve_by_attributes.return_value = [belief_data]

        def retrieve_by_id_side_effect(collection, id):
            if collection == "statements" and id == self.statement.id:
                return statement_data
            return None

        self.storage_adapter.retrieve_by_id.side_effect = retrieve_by_id_side_effect

        results = self.repository.search_by_notions(self.subject_notion, self.complement_notion)
        
        expected_query = {
            'subject_notion.id': "human_id",
            'complement_notion.id': "mortal_id"
        }
        self.storage_adapter.retrieve_by_attributes.assert_called_once_with("beliefs", expected_query)
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].id, self.belief.id)

    def test_search_by_subject_notion(self):
        belief_data = BeliefStorageMapper.to_dict(self.belief)
        statement_data = StatementStorageMapper.to_dict(self.statement)
        self.storage_adapter.retrieve_by_attributes.return_value = [belief_data]

        def retrieve_by_id_side_effect(collection, id):
            if collection == "statements" and id == self.statement.id:
                return statement_data
            return None

        self.storage_adapter.retrieve_by_id.side_effect = retrieve_by_id_side_effect

        results = self.repository.search_by_subject_notion(self.subject_notion)
        
        expected_query = {'subject_notion.id': "human_id"}
        self.storage_adapter.retrieve_by_attributes.assert_called_once_with("beliefs", expected_query)
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].id, self.belief.id)

    def test_search_by_complement_notion(self):
        belief_data = BeliefStorageMapper.to_dict(self.belief)
        statement_data = StatementStorageMapper.to_dict(self.statement)
        self.storage_adapter.retrieve_by_attributes.return_value = [belief_data]

        def retrieve_by_id_side_effect(collection, id):
            if collection == "statements" and id == self.statement.id:
                return statement_data
            return None

        self.storage_adapter.retrieve_by_id.side_effect = retrieve_by_id_side_effect

        results = self.repository.search_by_complement_notion(self.complement_notion)
        
        expected_query = {'complement_notion.id': "mortal_id"}
        self.storage_adapter.retrieve_by_attributes.assert_called_once_with("beliefs", expected_query)
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].id, self.belief.id)

if __name__ == '__main__':
    unittest.main()



from typing import Optional
from dataclasses import replace
from analogai.foundation.core.value_objects.modifier import Modifier
from analogai.foundation.core.value_objects.time import Time
from analogai.foundation.common.enums.deontic_modality import DeonticModality

class ModifierBuilder:
    
    def __init__(self):
        self._context = Modifier()
    
    def with_location(self, location: str):
        self._context.location = location
        return self
    
    def with_deontic_modality(self, modality: DeonticModality):
        self._context.deontic_modality = modality
        return self
    
    def with_from_time(self, year: Optional[int] = None, month: Optional[int] = None,
                      day: Optional[int] = None, hour: Optional[int] = None,
                      minute: Optional[int] = None):
        self._context.from_time = Time(year=year, month=month, day=day, hour=hour, minute=minute)
        return self
    
    def with_to_time(self, year: Optional[int] = None, month: Optional[int] = None,
                    day: Optional[int] = None, hour: Optional[int] = None,
                    minute: Optional[int] = None):
        self._context.to_time = Time(year=year, month=month, day=day, hour=hour, minute=minute)
        return self

    def with_attributes(self, attributes):
        self._context.attributes = attributes
        return self

    def with_quantity(self, quantity: Optional[float]):
        self._context.quantity = quantity
        return self

    def with_measures(self, measures):
        return self.with_attributes(measures)
    
    def build(self) -> Modifier:
        return replace(self._context)



import unittest
from unittest.mock import MagicMock

from analogai.foundation.common.design_patterns.state_machine import StateMachine

class TestStateMachine(unittest.TestCase):
    def setUp(self):
        self.context = MagicMock()
        self.state_machine = StateMachine(self.context)
        
        self.execute_result = MagicMock()
        self.mock_state_instance = MagicMock()
        self.mock_state_instance.execute.return_value = self.execute_result
        
        self.mock_state_class = MagicMock()
    
    def test_initialization(self):
        self.assertEqual(self.state_machine.context, self.context)
        self.assertIsNone(self.state_machine.current_state)
        self.assertIsNone(self.state_machine.previous_state)
    
    def test_state_to_first_transition(self):
        self.state_machine.state_to(self.mock_state_class)
        
        self.assertEqual(self.state_machine.current_state, self.mock_state_class)
        self.assertIsNone(self.state_machine.previous_state)
    
    def test_state_to_subsequent_transition(self):
        first_state = MagicMock()
        self.state_machine.current_state = first_state
        
        self.state_machine.state_to(self.mock_state_class)
        
        self.assertEqual(self.state_machine.current_state, self.mock_state_class)
        self.assertEqual(self.state_machine.previous_state, first_state)
    
    def test_set_context(self):
        new_context = MagicMock()
        self.state_machine.set_context(new_context)
        self.assertEqual(self.state_machine.context, new_context)

if __name__ == '__main__':
    unittest.main() 



from dataclasses import dataclass, field
from datetime import datetime
from typing import Optional
import uuid

@dataclass
class Turn:
    user_id: str
    agent_id: str
    user_message: str
    agent_response: str
    timestamp: datetime
    id: str = field(default_factory=lambda: str(uuid.uuid4()))



import unittest
from unittest.mock import Mock

from analogai.foundation.modules.categorical_deduction.categorical_deduction_service import CategoricalDeductionService
from analogai.foundation.modules.belief.belief import Belief
from analogai.foundation.modules.direct_statement.direct_statement import DirectStatement
from analogai.foundation.modules.notion.notion import Notion
from analogai.foundation.core.value_objects.relation import Relation


class FakeBeliefService:
    def __init__(self, beliefs):
        self._beliefs = list(beliefs)
        self.called_with = None

    def get_all_for_agent(self, agent_id: str):
        self.called_with = agent_id
        return list(self._beliefs)


class TestCategoricalDeductionService(unittest.TestCase):
    def setUp(self):
        self.agent_id = "agent-1"
        self.user_id = "user-1"

    def _build_belief(self, subject_id: str, object_id: str) -> Belief:
        subject = Notion(user_id=self.user_id, agent_id=self.agent_id, caption=subject_id, id=subject_id)
        complement_notion = Notion(user_id=self.user_id, agent_id=self.agent_id, caption=object_id, id=object_id)
        statement = DirectStatement(
            user_id=self.user_id,
            agent_id=self.agent_id,
            subject_notion=subject,
            complement_notion=complement_notion,
            relation=Relation.from_string("is"),
            probability=1.0,
            validity=1.0,
        )
        belief = Belief(agent_id=self.agent_id, id=f"belief-{subject_id}-{object_id}")
        belief.add_statement(statement)
        return belief

    def test_find_deductions_by_agent_id_uses_belief_service_and_paginates(self):
        belief_a = self._build_belief("a", "b")
        belief_b = self._build_belief("b", "c")
        belief_c = self._build_belief("c", "d")
        belief_service = FakeBeliefService([belief_a, belief_b, belief_c])
        service = CategoricalDeductionService(belief_service=belief_service)

        captured = {}

        def fake_resolve_deductions(*, beliefs_by_key, modifier, max_depth):
            captured["beliefs_by_key"] = beliefs_by_key
            captured["modifier"] = modifier
            captured["max_depth"] = max_depth
            return []

        service.resolve_deductions = fake_resolve_deductions

        service.find_deductions_by_agent_id(
            agent_id=self.agent_id,
            limit=1,
            offset=1,
            max_depth=3,
        )

        self.assertEqual(belief_service.called_with, self.agent_id)
        expected_key = service._belief_key_from_values(
            agent_id=belief_b.agent_id,
            subject_id=belief_b.subject_notion.id,
            object_id=belief_b.complement_notion.id,
            relation=belief_b.relation,
            modifier=belief_b.modifier,
        )
        self.assertEqual(set(captured["beliefs_by_key"].keys()), {expected_key})
        self.assertEqual(captured["max_depth"], 3)

    def test_find_deductions_by_agent_id_requires_belief_service(self):
        with self.assertRaises(ValueError):
            CategoricalDeductionService(belief_service=None)


if __name__ == "__main__":
    unittest.main()



from enum import Enum


class RelationshipType(Enum):
    IS_A = "is"
    DO = "do"
    RELATED_TO = "is related to"
    FORM_OF = "is form of"
    PART_OF = "is part of"
    HAS_A = "has"
    USED_FOR = "used for"
    CAPABLE_OF = "capable of"
    AT_LOCATION = "at location"
    CAUSES = "causes"
    HAS_SUBEVENT = "has subevent"
    HAS_FIRST_SUBEVENT = "has first subevent"
    HAS_LAST_SUBEVENT = "has last subevent"
    HAS_PREREQUISITE = "has prerequisite"
    HAS_PROPERTY = "has property"
    MOTIVATED_BY_GOAL = "motivated by goal"
    OBSTRUCTED_BY = "obstructed by"
    DESIRES = "desires"
    CREATED_BY = "created by"
    SYNONYM = "synonym"
    ANTONYM = "antonym"
    DISTINCT_FROM = "distinct from"
    DERIVED_FROM = "derived from"
    SYMBOL_OF = "symbol of"
    DEFINED_AS = "defined as"
    MANNER_OF = "manner of"
    LOCATED_NEAR = "located near"
    HAS_CONTEXT = "has context"
    SIMILAR_TO = "similar to"
    ETYMOLOGICALLY_RELATED_TO = "etymologically related to"
    ETYMOLOGICALLY_DERIVED_FROM = "etymologically derived from"
    CAUSES_DESIRE = "causes desire"
    MADE_OF = "made of"
    RECEIVES_ACTION = "receives action"



from abc import ABC, abstractmethod
from typing import List, Optional
from analogai.foundation.extensions.conversation.conversation import Conversation
from analogai.foundation.extensions.conversation.turn import Turn


class ConversationRepository(ABC):
    @abstractmethod
    def create_conversation(self, user_id: str, agent_id: str) -> Conversation:
        pass

    @abstractmethod
    def find_by_id(self, conversation_id: str) -> Optional[Conversation]:
        pass

    @abstractmethod
    def find_by_agent(self, agent_id: str, page: int = 1, page_size: int = 10) -> List[Conversation]:
        pass

    @abstractmethod
    def find_latest_by_user_and_agent(self, user_id: str, agent_id: str) -> Optional[Conversation]:
        pass

    @abstractmethod
    def add_turn(self, conversation_id: str, turn: Turn) -> Turn:
        pass

    @abstractmethod
    def delete_conversation(self, conversation_id: str) -> bool:
        pass



"""Embedding providers and factory helpers."""

from .base import EmbeddingProvider
from .provider_factory import create_embedding_provider

__all__ = [
    "EmbeddingProvider",
    "create_embedding_provider",
]



from typing import Iterable, List, Optional, Dict, Callable
from datetime import datetime

from analogai.foundation.modules.belief.belief import Belief
from analogai.foundation.modules.shared.statement_base import StatementBase


class BeliefAssembler:
    def build_belief(
        self,
        statements: Iterable[StatementBase],
        belief_id: Optional[str] = None,
    ) -> Belief:
        statements_list = list(statements)
        if not statements_list:
            raise ValueError("Cannot build belief without statements")
        ordered = sorted(
            statements_list,
            key=lambda stmt: getattr(stmt, "created_at", datetime.min),
        )
        first = ordered[0]
        belief = Belief(
            agent_id=first.agent_id,
            id=belief_id or first.id,
        )
        for statement in ordered:
            belief.add_statement(statement)
        created_candidates: List[datetime] = []
        for stmt in ordered:
            created_at = getattr(stmt, "created_at", None)
            if isinstance(created_at, datetime):
                created_candidates.append(created_at)
        created_at = min(created_candidates) if created_candidates else None
        updated_at = max(created_candidates) if created_candidates else None
        if created_at:
            belief.created_at = created_at
        if updated_at:
            belief.updated_at = updated_at
        return belief

    def group_statements(
        self,
        statements: Iterable[StatementBase],
        key_builder: Callable[[StatementBase], object],
    ) -> Dict[object, List[StatementBase]]:
        groups: Dict[object, List[StatementBase]] = {}
        for statement in statements:
            key = key_builder(statement)
            groups.setdefault(key, []).append(statement)
        return groups



from typing import Optional, List
from datetime import datetime
import uuid
from analogai.foundation.extensions.conversation.repositories import ConversationRepository
from analogai.foundation.extensions.conversation.conversation import Conversation
from analogai.foundation.extensions.conversation.turn import Turn
from analogai.foundation.modules.user.user import User
from analogai.foundation.modules.agent.agent import Agent
from analogai.foundation.common.logging.app_logger import app_logger
from typeguard import typechecked


class ConversationService:
    def __init__(self, conversation_repository: ConversationRepository):
        self.repository = conversation_repository

    @typechecked
    def start_conversation(self, user_id: str, agent_id: str) -> Conversation:
        conversation = self.repository.create_conversation(user_id, agent_id)
        app_logger.debug(f"Created Conversation: {conversation.id}")
        return conversation

    @typechecked
    def add_turn(
        self,
        conversation_id: str,
        user_id: str,
        agent_id: str,
        user_message: str,
        agent_response: str,
    ) -> Turn:
        turn_id = str(uuid.uuid4())
        turn = Turn(
            id=turn_id,
            user_id=user_id,
            agent_id=agent_id,
            user_message=user_message,
            agent_response=agent_response,
            timestamp=datetime.now(),
        )

        self.repository.add_turn(conversation_id, turn)
        app_logger.debug(f"Added Turn to Conversation (turn_id: {turn.id}, conversation_id: {conversation_id})")
        return turn

    @typechecked
    def find(self, conversation_id: str) -> Optional[Conversation]:
        result = self.repository.find_by_id(conversation_id)
        if result:
            app_logger.debug(f"Found Conversation (conversation_id: {result.id})")
        else:
            app_logger.debug(f"Conversation not found: id={conversation_id}")
        return result

    @typechecked
    def get_conversations_by_agent_id(self, agent_id: str, page: int = 1, page_size: int = 10) -> List[Conversation]:
        return self.repository.find_by_agent(agent_id, page, page_size)

    @typechecked
    def delete_conversation(self, conversation_id: str) -> bool:
        success = self.repository.delete_conversation(conversation_id)
        if success:
            app_logger.debug(f"Deleted Conversation {conversation_id}")
        return success

    @typechecked
    def get_latest_conversation(self, user_id: str, agent_id: str) -> Optional[Conversation]:
        result = self.repository.find_latest_by_user_and_agent(user_id, agent_id)
        if result:
            app_logger.debug(f"Found latest Conversation (conversation_id: {result.id})")
        else:
            app_logger.debug(f"Latest conversation not found")
        return result

    @typechecked
    def ensure_conversation_exists(self, user_id: str, agent_id: str) -> Conversation:
        existing_conversation = self.repository.find_latest_by_user_and_agent(user_id, agent_id)
        if existing_conversation:
            app_logger.debug(f"Found existing Conversation (conversation_id: {existing_conversation.id})")
            return existing_conversation

        conversation = self.repository.create_conversation(user_id, agent_id)
        app_logger.debug(f"Created Conversation (conversation_id: {conversation.id})")
        return conversation



import unittest

from analogai.foundation.common.enums.relationship_type import RelationshipType
from analogai.foundation.core.value_objects.relation import Relation


class TestRelation(unittest.TestCase):
	def test_relation_equality_with_same_value(self):
		self.assertEqual(Relation.from_string("is"), Relation.from_string("is"))

	def test_relation_inequality_with_different_value(self):
		self.assertNotEqual(Relation.from_string("is"), Relation.from_string("has"))

	def test_is_of_type_with_known_relationship_type(self):
		relation = Relation.from_type(RelationshipType.CAUSES)

		self.assertTrue(relation.is_of_type(RelationshipType.CAUSES))
		self.assertFalse(relation.is_of_type(RelationshipType.IS_A))

	def test_is_of_type_with_custom_relation_value(self):
		relation = Relation.from_string("sell")

		self.assertFalse(relation.is_of_type(RelationshipType.IS_A))
		self.assertFalse(relation.is_of_type(RelationshipType.CAUSES))




if __name__ == '__main__':
	unittest.main()



from typing import Dict, Optional, Callable

from analogai.foundation.modules.belief.belief import Belief
from analogai.foundation.modules.shared.statement_base import StatementBase
from analogai.foundation.modules.belief.belief_filterer import BeliefKey
from analogai.foundation.core.value_objects.modifier import Modifier


class BeliefDeductionApplier:
    def __init__(
        self,
        belief_assembler,
        belief_key_builder: Callable[[StatementBase], BeliefKey],
        categorical_deduction_service=None,
        conditional_deduction_service=None,
        max_deduction_depth: int = 5,
    ):
        self._belief_assembler = belief_assembler
        self._belief_key_builder = belief_key_builder
        self._categorical_deduction_service = categorical_deduction_service
        self._conditional_deduction_service = conditional_deduction_service
        self._max_deduction_depth = max_deduction_depth

    def set_categorical_deduction_service(self, service) -> None:
        self._categorical_deduction_service = service

    def set_conditional_deduction_service(self, service) -> None:
        self._conditional_deduction_service = service

    def apply_deductions(
        self,
        beliefs_by_key: Dict[BeliefKey, Belief],
        agent_id: str,
        modifier: Optional[Modifier],
        quantity: Optional[float],
        user_id: Optional[str] = None,
    ) -> Dict[BeliefKey, Belief]:
        if self._categorical_deduction_service is not None:
            self._categorical_deduction_service.resolve_deductions(
                beliefs_by_key,
                modifier=modifier,
                max_depth=self._max_deduction_depth,
            )

        if self._conditional_deduction_service is not None:
            deductions = self._conditional_deduction_service.resolve_deductions(
                agent_id=agent_id,
                beliefs=list(beliefs_by_key.values()),
                modifier=modifier,
                user_id=user_id,
            )
            for deduction in deductions:
                belief_key = self._belief_key_builder(deduction)
                belief = beliefs_by_key.get(belief_key)
                if belief is None:
                    beliefs_by_key[belief_key] = self._belief_assembler.build_belief([deduction])
                else:
                    belief.add_statement(deduction)

        return beliefs_by_key



from typing import TYPE_CHECKING

__all__ = [
    "AuthorizationOperation",
    "AuthorizationService",
    "AuthorizationGuard",
    "AuthorizationGuardError",
    "AuthorizationErrorReason",
    "AuthorizationFactoryInterface",
    "AuthorizationContext",
    "AuthorizationFactory",
    "build_authorization_dependency",
    "build_api_authorization_dependency",
]

if TYPE_CHECKING:
    from .core import (
        AuthorizationOperation,
        AuthorizationService,
        AuthorizationGuard,
        AuthorizationGuardError,
        AuthorizationErrorReason,
    )
    from .interfaces import AuthorizationFactoryInterface
    from .models import AuthorizationContext
    from .authorization_factory import AuthorizationFactory
    from .adapters.fastapi import build_authorization_dependency, build_api_authorization_dependency

def __getattr__(name: str):
    if name in {
        "AuthorizationOperation",
        "AuthorizationService",
        "AuthorizationGuard",
        "AuthorizationGuardError",
        "AuthorizationErrorReason",
    }:
        from . import core
        return getattr(core, name)
    if name == "AuthorizationFactoryInterface":
        from .interfaces import AuthorizationFactoryInterface
        return AuthorizationFactoryInterface
    if name == "AuthorizationContext":
        from .models import AuthorizationContext
        return AuthorizationContext
    if name == "AuthorizationFactory":
        from .authorization_factory import AuthorizationFactory
        return AuthorizationFactory
    if name == "build_authorization_dependency":
        from .adapters.fastapi import build_authorization_dependency
        return build_authorization_dependency
    if name == "build_api_authorization_dependency":
        from .adapters.fastapi import build_api_authorization_dependency
        return build_api_authorization_dependency
    raise AttributeError(name)



from analogai.foundation.common.enums.environment import Environment
from analogai.foundation.settings import config
from analogai.foundation.infrastructure.factories import StorageAdaptersFactoryImpl
# Application-layer factories removed. Use direct mock or DI as needed.
from analogai.foundation.tests.functional.storage_cleaner import cleanup_storages
from analogai.foundation.modules.user.user import User
from analogai.foundation.modules.agent.agent import Agent, Role
import uuid

from analogai.foundation.setup.factories import (
    AgentServiceFactory,
    BeliefServiceFactory,
    CategoricalDeductionServiceFactory,
    ConditionalDeductionServiceFactory,
    HypotheticalStatementServiceFactory,
    NotionServiceFactory,
    StatementServiceFactory,
    UserServiceFactory,
    create_repository_strategy,
)


class ServicesTestConfig:
    def __init__(self):
        self._config_override = config.override(**{"vector.enabled": False})
        self._config_override.__enter__()
        self.storage_adapters_factory = StorageAdaptersFactoryImpl()
        self.storage_adapter = self.storage_adapters_factory.get_storage_adapter()

        self._env = Environment(config.get("ENVIRONMENT"))
        self.repository_strategy = create_repository_strategy()

        self.user = User(id=str(uuid.uuid4()), name="Test User", email="test@example.com")
        self.agent = Agent(id=str(uuid.uuid4()), creator=self.user, roles=[Role(user=self.user, authority=1.0)])

        self.user_repository = self.repository_strategy.get_user_repository()
        self.agent_repository = self.repository_strategy.get_agent_repository()
        self.notion_repository = self.repository_strategy.get_notion_repository()
        self.statement_repository = self.repository_strategy.get_statement_repository()
        self.belief_repository = self.repository_strategy.get_belief_repository()
        self.hypothetical_statement_repository = self.repository_strategy.get_hypothetical_statement_repository()

        self.user_repository.create(self.user)
        self.agent_repository.create(self.agent)

        self.user_service = UserServiceFactory(repository=self.user_repository).get_service()
        self.agent_service = AgentServiceFactory(repository=self.agent_repository).get_service()
        self.notion_service = NotionServiceFactory(repository=self.notion_repository).get_service()
        self.statement_service = StatementServiceFactory(repository=self.statement_repository).get_service()
        self.belief_service = BeliefServiceFactory(
            repository=self.belief_repository,
            notion_service=self.notion_service,
            statement_service=self.statement_service,
        ).get_service()
        self.categorical_deduction_service = CategoricalDeductionServiceFactory(
            belief_service=self.belief_service,
        ).get_service()
        self.conditional_deduction_service = ConditionalDeductionServiceFactory(
            belief_service=self.belief_service,
        ).get_service()
        self.hypothetical_statement_service = HypotheticalStatementServiceFactory(
            repository=self.hypothetical_statement_repository,
        ).get_service()

        self.belief_service.set_categorical_deduction_service(self.categorical_deduction_service)
        self.belief_service.set_conditional_deduction_service(self.conditional_deduction_service)

    def tear_down(self):
        if self._config_override:
            self._config_override.__exit__(None, None, None)
        cleanup_storages()



from analogai.foundation.infrastructure.storage.storage_adapters_factory import StorageAdaptersFactory
from analogai.foundation.infrastructure.storage.in_memory.in_memory_storage_adapter_impl import InMemoryStorageAdapterImpl
# from analogai.foundation.infrastructure.storage.in_memory.in_memory_query_storage_adapter_impl import InMemoryQueryStorageAdapterImpl

class InMemoryStorageAdaptersStrategy(StorageAdaptersFactory):
    def get_storage_adapter(self):
        return InMemoryStorageAdapterImpl()

    # def get_query_storage_adapter(self):
    #     return InMemoryQueryStorageAdapterImpl()




from __future__ import annotations

import json
from typing import Optional
from fastapi import HTTPException, Request, status
from analogai.foundation.modules.user.user import User
from analogai.foundation.modules.user.user_service import UserService
from analogai.foundation.extensions.authentication.interfaces.authentication_factory_interface import (
	AuthenticationDependency,
	AuthenticationFactoryInterface,
	TokenExtractionDependency,
	ApiTokenAuthenticationDependency,
)
from analogai.foundation.extensions.authentication.core import IJWTService
from analogai.foundation.extensions.authentication.models import TokenPayload
from analogai.foundation.common.logging.app_logger import app_logger
from analogai.foundation.extensions.authentication.api_token.api_token_authentication_service import (
	ApiTokenAuthenticationService,
)
from analogai.foundation.extensions.authentication.api_token.auth_type_handlers import (
	ApiAuthTypeHandler,
	GuestApiAuthHandler,
	SystemApiAuthHandler,
	UserApiAuthHandler,
)
from analogai.foundation.extensions.authentication.providers.auth0 import Auth0JWTService
from analogai.foundation.setup.factories.repository_selector import create_repository_strategy


class AuthenticationFactory(AuthenticationFactoryInterface):
	def __init__(self):
		self._jwt_service: Optional[IJWTService] = None
		self._user_service: Optional[UserService] = None
		self._api_token_service: Optional[ApiTokenAuthenticationService] = None
		self._api_auth_handlers: Optional[dict[str, ApiAuthTypeHandler]] = None

	def create_authentication_dependency(self) -> AuthenticationDependency:
		async def dependency(request: Request) -> User:
			token = self._extract_token(request)
			try:
				payload = self._get_jwt_service().decode_payload(token)
			except ValueError as error:
				raise HTTPException(
					status_code=status.HTTP_401_UNAUTHORIZED,
					detail=str(error),
				) from error

			if payload.get("email"):
				try:
					user = self._get_jwt_service().authenticate_user(token)
				except ValueError as error:
					detail = str(error)
					if "user not found" in detail.lower():
						detail = "User not found for token"
					raise HTTPException(
						status_code=status.HTTP_401_UNAUTHORIZED,
						detail=detail,
					) from error
			else:
				user = await self._authenticate_api_token(request, token)

			if not user:
				raise HTTPException(
					status_code=status.HTTP_401_UNAUTHORIZED,
					detail="User not found for token",
				)

			return user

		return dependency

	def create_token_extraction_dependency(self) -> TokenExtractionDependency:
		async def dependency(request: Request) -> TokenPayload:
			token = self._extract_token(request)
			try:
				return self._get_jwt_service().extract_payload(token)
			except ValueError as error:
				raise HTTPException(
					status_code=status.HTTP_401_UNAUTHORIZED,
					detail=str(error),
				) from error

		return dependency

	def create_api_token_authentication_dependency(self) -> ApiTokenAuthenticationDependency:
		async def dependency(request: Request) -> User:
			token = self._extract_token(request)
			return await self._authenticate_api_token(request, token)

		return dependency

	async def _authenticate_api_token(self, request: Request, token: str) -> User:
		try:
			api_payload = self._get_api_token_service().extract_payload(token)
		except ValueError as error:
			raise HTTPException(
				status_code=status.HTTP_401_UNAUTHORIZED,
				detail=str(error),
			) from error

		auth_payload = await self._extract_api_auth_payload(request)
		auth_type = str(auth_payload.get("auth_type", "system")).lower()

		handler = self._get_api_auth_handlers().get(auth_type)
		if not handler:
			raise HTTPException(
				status_code=status.HTTP_401_UNAUTHORIZED,
				detail="Invalid auth_type; expected 'user', 'system', or 'guest'",
			)

		try:
			user = handler.authenticate(
				auth_payload=auth_payload,
				api_payload=api_payload,
				user_service=self._get_user_service(),
			)
		except ValueError as error:
			raise HTTPException(
				status_code=status.HTTP_401_UNAUTHORIZED,
				detail=str(error),
			) from error

		request.state.api_auth_type = auth_type
		request.state.api_agent_id = api_payload.agent_id
		request.state.api_user_email = user.email if user else None
		return user

	def _get_jwt_service(self) -> IJWTService:
		if not self._jwt_service:
			from analogai.foundation.extensions.authentication import authentication_factory as auth_factory

			self._jwt_service = auth_factory.Auth0JWTService(self._get_user_service())
		return self._jwt_service

	def _get_user_service(self) -> UserService:
		if not self._user_service:
			from analogai.foundation.extensions.authentication import authentication_factory as auth_factory

			repository = auth_factory.create_repository_strategy().get_user_repository()
			self._user_service = UserService(repository)
		return self._user_service

	def _get_api_token_service(self) -> ApiTokenAuthenticationService:
		if not self._api_token_service:
			self._api_token_service = ApiTokenAuthenticationService()
		return self._api_token_service

	def _get_api_auth_handlers(self) -> dict[str, ApiAuthTypeHandler]:
		if not self._api_auth_handlers:
			handlers = (
				UserApiAuthHandler(),
				SystemApiAuthHandler(),
				GuestApiAuthHandler(),
			)
			self._api_auth_handlers = {handler.auth_type: handler for handler in handlers}
		return self._api_auth_handlers

	async def _extract_api_auth_payload(self, request: Request) -> dict:
		body = await request.body()
		if not body:
			return {}
		payload = json.loads(body)
		if isinstance(payload, dict):
			return payload
		return {}

	def _extract_token(self, request: Request) -> str:
		authorization = request.headers.get("Authorization", "")
		if not authorization.startswith("Bearer "):
			app_logger.debug(
				"Invalid Authorization header format: '%s'",
				authorization,
			)
			raise HTTPException(
				status_code=status.HTTP_401_UNAUTHORIZED,
				detail="Invalid authentication token format",
			)
		return authorization.replace("Bearer ", "", 1)


__all__ = ["AuthenticationFactory", "Auth0JWTService", "create_repository_strategy"]



from typing import Dict
from abc import ABC, abstractmethod
from analogai.foundation.settings import config
from analogai.foundation.common.utils.statement_serializers.helpers.relation_verbalization_map import RELATION_TO_VERBALIZATION


class VerbSelectionStrategy(ABC):
    @abstractmethod
    def select_verb(self, relation: str, probability: float, is_plural: bool) -> str:
        pass


class DefaultVerbSelectionStrategy(VerbSelectionStrategy):
    def __init__(self, verbalization_map: Dict):
        self.verbalization_map = verbalization_map
    
    def select_verb(self, relation: str, probability: float, is_plural: bool) -> str:
        verbalization = RELATION_TO_VERBALIZATION.get(relation)
        if verbalization is None:
            return relation
        try:
            forms = self.verbalization_map[verbalization.name]._forms(is_plural)
        except KeyError:
            return relation
        if probability < config.get("relation.pos.neg", 0.3):
            return forms["negative"]
        elif probability <= config.get("relation.pos.uncertain", 0.7):
            return forms["uncertain"]
        return forms["positive"]


class VerbSelector:
    def __init__(self, strategy: VerbSelectionStrategy):
        self.strategy = strategy
    
    def select(self, relation: str, probability: float, is_plural: bool) -> str:
        return self.strategy.select_verb(relation, probability, is_plural)


__all__ = ["VerbSelector", "DefaultVerbSelectionStrategy", "VerbSelectionStrategy"]



from .hypothetical_statement import HypotheticalStatement
from .hypothetical_statement_repository import HypotheticalStatementRepository
from .hypothetical_statement_service import HypotheticalStatementService
from .conditional_deduction import ConditionalDeduction
from .conditional_deduction_service import ConditionalDeductionService

__all__ = [
    "HypotheticalStatement",
    "HypotheticalStatementRepository",
    "HypotheticalStatementService",
    "ConditionalDeduction",
    "ConditionalDeductionService",
]



import unittest
from datetime import datetime
from analogai.foundation.extensions.fallback import Fallback


class TestFallback(unittest.TestCase):
    def setUp(self):
        self.user_id = "user-123"
        self.agent_id = "agent-456"
        self.user_question = "What is the meaning of life?"

    def test_create_fallback(self):
        fallback = Fallback(
            user_id=self.user_id,
            agent_id=self.agent_id,
            user_question=self.user_question
        )
        self.assertEqual(fallback.user_id, self.user_id)
        self.assertEqual(fallback.agent_id, self.agent_id)
        self.assertEqual(fallback.user_question, self.user_question)
        self.assertIsNotNone(fallback.id)
        self.assertIsNotNone(fallback.timestamp)
        self.assertIsInstance(fallback.timestamp, datetime)

    def test_fallback_has_unique_id(self):
        fallback1 = Fallback(
            user_id=self.user_id,
            agent_id=self.agent_id,
            user_question=self.user_question
        )
        fallback2 = Fallback(
            user_id=self.user_id,
            agent_id=self.agent_id,
            user_question=self.user_question
        )
        self.assertNotEqual(fallback1.id, fallback2.id)

    def test_fallback_can_have_custom_id(self):
        custom_id = "custom-fallback-id"
        fallback = Fallback(
            user_id=self.user_id,
            agent_id=self.agent_id,
            user_question=self.user_question,
            id=custom_id
        )
        self.assertEqual(fallback.id, custom_id)

    def test_fallback_can_have_custom_timestamp(self):
        custom_timestamp = datetime(2023, 1, 1, 12, 0, 0)
        fallback = Fallback(
            user_id=self.user_id,
            agent_id=self.agent_id,
            user_question=self.user_question,
            timestamp=custom_timestamp
        )
        self.assertEqual(fallback.timestamp, custom_timestamp)

    def test_fallback_validation_empty_user_id(self):
        with self.assertRaises(ValueError):
            Fallback(
                user_id="",
                agent_id=self.agent_id,
                user_question=self.user_question
            )

    def test_fallback_validation_empty_agent_id(self):
        with self.assertRaises(ValueError):
            Fallback(
                user_id=self.user_id,
                agent_id="",
                user_question=self.user_question
            )

    def test_fallback_validation_empty_user_question(self):
        with self.assertRaises(ValueError):
            Fallback(
                user_id=self.user_id,
                agent_id=self.agent_id,
                user_question=""
            )

    def test_fallback_validation_none_user_id(self):
        with self.assertRaises(ValueError):
            Fallback(
                user_id=None,
                agent_id=self.agent_id,
                user_question=self.user_question
            )


if __name__ == '__main__':
    unittest.main()



from abc import ABC, abstractmethod
from typing import Optional, List
from analogai.foundation.extensions.fallback import Fallback


class FallbackRepository(ABC):
    @abstractmethod
    def create(self, fallback: Fallback) -> Fallback:
        pass

    @abstractmethod
    def find_by_id(self, fallback_id: str) -> Optional[Fallback]:
        pass

    @abstractmethod
    def find_by_agent_id(self, agent_id: str) -> List[Fallback]:
        pass



from __future__ import annotations

import os
from importlib import resources

from dynaconf import Dynaconf

from analogai.foundation.common.utils.config_base import ConfigBase
from analogai.foundation.common.utils.env_utils import load_envs_in_order


load_envs_in_order()
if not os.getenv("ENVIRONMENT"):
    os.environ["ENVIRONMENT"] = "development"

_CONFIG_DIR = resources.files("analogai.foundation.config")
_DEFAULT_CONFIG = _CONFIG_DIR / "default.yaml"
_SCHEMA_CONFIG = _CONFIG_DIR / "schema.yaml"


settings = Dynaconf(
    envvar_prefix=False,
    settings_files=[
        str(_DEFAULT_CONFIG),
    ],
    defaults={
        "ENVIRONMENT": "development",
    },
    environments=False,
    load_dotenv=True,
    validate_on_update=True,
)
config = ConfigBase(settings, schema_path=_SCHEMA_CONFIG)

__all__ = ["settings", "config"]



from analogai.foundation.modules.belief.belief_service import BeliefService
from .notion_service_factory import NotionServiceFactory
from .statement_service_factory import StatementServiceFactory


class BeliefServiceFactory:
    def __init__(
        self,
        repository=None,
        notion_service=None,
        statement_service=None,
        categorical_deduction_service=None,
        conditional_deduction_service=None,
        environment: str | None = None,
    ):
        self.repository = repository
        self.statement_service = statement_service or StatementServiceFactory(environment=environment).get_service()
        self.notion_service = notion_service or NotionServiceFactory(environment=environment).get_service()
        self.service = BeliefService(
            statement_service=self.statement_service,
            notion_service=self.notion_service,
            categorical_deduction_service=categorical_deduction_service,
            conditional_deduction_service=conditional_deduction_service,
        )

    def get_service(self) -> BeliefService:
        return self.service

    def get_repository(self):
        return self.repository



import unittest
from analogai.foundation.extensions.conversation.factories import ConversationServiceFactory
from analogai.foundation.tests.functional.services.setup import ServicesTestConfig


class TestConversationServiceFunctional(unittest.TestCase):
    def setUp(self):
        self.config = ServicesTestConfig()
        self.conversation_service = ConversationServiceFactory().create()
        self.test_user = self.config.user
        self.test_agent = self.config.agent

    def tearDown(self):
        self.config.tear_down()

    def test_start_and_add_turns_to_conversation(self):
        conversation = self.conversation_service.start_conversation(self.test_user.id, self.test_agent.id)

        self.assertIsNotNone(conversation.id)
        self.assertEqual(conversation.user.id, self.test_user.id)
        self.assertEqual(conversation.agent.id, self.test_agent.id)

        turn = self.conversation_service.add_turn(
            conversation.id, self.test_user.id, self.test_agent.id,
            "Hello", "Hi there"
        )

        self.assertIsNotNone(turn.id)
        self.assertEqual(turn.user_id, self.test_user.id)
        self.assertEqual(turn.agent_id, self.test_agent.id)
        self.assertEqual(turn.user_message, "Hello")
        self.assertEqual(turn.agent_response, "Hi there")

    def test_get_conversations_by_agent(self):
        conv1 = self.conversation_service.start_conversation(self.test_user.id, self.test_agent.id)
        conv2 = self.conversation_service.start_conversation(self.test_user.id, self.test_agent.id)

        conversations = self.conversation_service.get_conversations_by_agent_id(self.test_agent.id)

        self.assertEqual(len(conversations), 2)
        self.assertIn(conv1.id, [c.id for c in conversations])
        self.assertIn(conv2.id, [c.id for c in conversations])

    def test_get_recent_turns_from_conversation(self):
        conversation = self.conversation_service.start_conversation(self.test_user.id, self.test_agent.id)
        self.conversation_service.add_turn(conversation.id, self.test_user.id, self.test_agent.id, "Message 1", "Response 1")
        self.conversation_service.add_turn(conversation.id, self.test_user.id, self.test_agent.id, "Message 2", "Response 2")

        retrieved_conversation = self.conversation_service.find(conversation.id)
        recent_turns = retrieved_conversation.get_recent_turns(limit=10)

        self.assertEqual(len(recent_turns), 2)
        self.assertEqual(recent_turns[0].user_message, "Message 1")
        self.assertEqual(recent_turns[1].user_message, "Message 2")

    def test_get_latest_turn_from_conversation(self):
        conversation = self.conversation_service.start_conversation(self.test_user.id, self.test_agent.id)
        self.conversation_service.add_turn(conversation.id, self.test_user.id, self.test_agent.id, "Message 1", "Response 1")
        self.conversation_service.add_turn(conversation.id, self.test_user.id, self.test_agent.id, "Message 2", "Response 2")

        retrieved_conversation = self.conversation_service.find(conversation.id)
        latest_turn = retrieved_conversation.get_latest_turn()

        self.assertIsNotNone(latest_turn)
        self.assertEqual(latest_turn.user_message, "Message 2")
        self.assertEqual(latest_turn.agent_response, "Response 2")

    def test_delete_conversation(self):
        conversation = self.conversation_service.start_conversation(self.test_user.id, self.test_agent.id)

        success = self.conversation_service.delete_conversation(conversation.id)

        self.assertTrue(success)
        retrieved = self.conversation_service.find(conversation.id)
        self.assertIsNone(retrieved)

    def test_ensure_conversation_exists_returns_existing_conversation(self):
        conversation = self.conversation_service.start_conversation(self.test_user.id, self.test_agent.id)

        ensured_conversation = self.conversation_service.ensure_conversation_exists(self.test_user.id, self.test_agent.id)

        self.assertIsNotNone(ensured_conversation)
        self.assertEqual(ensured_conversation.id, conversation.id)
        self.assertEqual(ensured_conversation.user.id, self.test_user.id)
        self.assertEqual(ensured_conversation.agent.id, self.test_agent.id)

    def test_ensure_conversation_exists_creates_new_conversation_when_none_exists(self):
        ensured_conversation = self.conversation_service.ensure_conversation_exists(self.test_user.id, self.test_agent.id)

        self.assertIsNotNone(ensured_conversation)
        self.assertIsNotNone(ensured_conversation.id)
        self.assertEqual(ensured_conversation.user.id, self.test_user.id)
        self.assertEqual(ensured_conversation.agent.id, self.test_agent.id)

        retrieved = self.conversation_service.find(ensured_conversation.id)
        self.assertIsNotNone(retrieved)
        self.assertEqual(retrieved.id, ensured_conversation.id)


if __name__ == '__main__':
    unittest.main()



import unittest
from datetime import datetime
from analogai.foundation.modules.notion.notion import Notion
from analogai.foundation.core.value_objects.modifier import Modifier
from analogai.foundation.core.value_objects.relation import Relation
from analogai.foundation.common.utils.modifier_utils import ModifierBuilder
from analogai.foundation.common.enums.deontic_modality import DeonticModality
from analogai.foundation.tests.functional.services.setup import ServicesTestConfig


class TestStatementServiceFunctional(unittest.TestCase):
    def setUp(self):
        self.config = ServicesTestConfig()
        self.statement_service = self.config.statement_service
        self.user = self.config.user
        self.agent = self.config.agent
        self.subject_notion = Notion(
            user_id=self.user.id,
            agent_id=self.agent.id,
            caption="socrates",
            id="socrates_id",
        )
        self.relation = Relation.from_string("sell")
        self.complement_notion = Notion(
            user_id=self.user.id,
            agent_id=self.agent.id,
            caption="iphone",
            id="iphone_id",
        )
        self.config.notion_repository.create(self.subject_notion)
        self.config.notion_repository.create(self.complement_notion)

    def tearDown(self):
        self.config.tear_down()

    def test_create_statement_with_quantity_socrates_sell_2_iphones(self):
        created = self.statement_service.create(
            user_id=self.user.id,
            agent_id=self.agent.id,
            subject_notion=self.subject_notion,
            relation=self.relation,
            complement_notion=self.complement_notion,
            quantity=2.0,
        )
        self.assertIsNotNone(created.id)
        self.assertEqual(created.quantity, 2.0)
        self.assertEqual(created.subject_notion.caption, "socrates")
        self.assertEqual(created.relation, Relation.from_string("sell"))
        self.assertEqual(created.complement_notion.caption, "iphone")

        found = self.statement_service.find_by_id(created.id)
        self.assertIsNotNone(found)
        self.assertEqual(found.quantity, 2.0)
        self.assertIsNotNone(found.created_at)
        self.assertIsInstance(found.created_at, datetime)

    def test_statement_created_at_persists(self):
        created = self.statement_service.create(
            user_id=self.user.id,
            agent_id=self.agent.id,
            subject_notion=self.subject_notion,
            relation=self.relation,
            complement_notion=self.complement_notion,
        )
        self.assertIsNotNone(created.created_at)
        original_created_at = created.created_at

        found = self.statement_service.find_by_id(created.id)
        self.assertIsNotNone(found)
        self.assertIsNotNone(found.created_at)
        self.assertEqual(found.created_at.isoformat(), original_created_at.isoformat())

    def test_statement_modifier_persists_location_deontic_and_times(self):
        modifier = ModifierBuilder().with_location("Paris").with_deontic_modality(
            DeonticModality.OBLIGATORY
        ).with_from_time(year=2024, month=3, day=15, hour=10, minute=30).with_to_time(
            year=2024, month=6, day=20
        ).build()
        created = self.statement_service.create(
            user_id=self.user.id,
            agent_id=self.agent.id,
            subject_notion=self.subject_notion,
            relation=self.relation,
            complement_notion=self.complement_notion,
            modifier=modifier,
        )
        self.assertIsNotNone(created.id)
        self.assertIsNotNone(created.modifier)
        self.assertEqual(created.modifier.location, "Paris")
        self.assertEqual(created.modifier.deontic_modality, DeonticModality.OBLIGATORY)
        self.assertEqual(created.modifier.from_time.year, 2024)
        self.assertEqual(created.modifier.from_time.month, 3)
        self.assertEqual(created.modifier.from_time.day, 15)
        self.assertEqual(created.modifier.from_time.hour, 10)
        self.assertEqual(created.modifier.from_time.minute, 30)
        self.assertEqual(created.modifier.to_time.year, 2024)
        self.assertEqual(created.modifier.to_time.month, 6)
        self.assertEqual(created.modifier.to_time.day, 20)

        found = self.statement_service.find_by_id(created.id)
        self.assertIsNotNone(found)
        self.assertIsNotNone(found.modifier)
        self.assertEqual(found.modifier.location, "Paris")
        self.assertEqual(found.modifier.deontic_modality, DeonticModality.OBLIGATORY)
        self.assertEqual(found.modifier.from_time.year, 2024)
        self.assertEqual(found.modifier.from_time.month, 3)
        self.assertEqual(found.modifier.from_time.day, 15)
        self.assertEqual(found.modifier.from_time.hour, 10)
        self.assertEqual(found.modifier.from_time.minute, 30)
        self.assertEqual(found.modifier.to_time.year, 2024)
        self.assertEqual(found.modifier.to_time.month, 6)
        self.assertEqual(found.modifier.to_time.day, 20)

    def test_statement_find_by_subject_and_relation_filters_by_modifier(self):
        modifier_paris = Modifier(location="Paris", deontic_modality=DeonticModality.OBLIGATORY)
        modifier_london = Modifier(location="London", deontic_modality=DeonticModality.PERMITTED)
        self.statement_service.create(
            user_id=self.user.id,
            agent_id=self.agent.id,
            subject_notion=self.subject_notion,
            relation=self.relation,
            complement_notion=self.complement_notion,
            modifier=modifier_paris,
        )
        self.statement_service.create(
            user_id=self.user.id,
            agent_id=self.agent.id,
            subject_notion=self.subject_notion,
            relation=self.relation,
            complement_notion=self.complement_notion,
            modifier=modifier_london,
        )

        results_paris = self.statement_service.find_by_subject_and_relation(
            self.subject_notion, Relation.from_string("sell"), modifier=Modifier(location="Paris")
        )
        results_london = self.statement_service.find_by_subject_and_relation(
            self.subject_notion, Relation.from_string("sell"), modifier=Modifier(location="London")
        )
        self.assertEqual(len(results_paris), 1)
        self.assertEqual(len(results_london), 1)
        self.assertEqual(results_paris[0].modifier.location, "Paris")
        self.assertEqual(results_london[0].modifier.location, "London")

    def test_statement_find_by_object_and_relation_filters_by_modifier(self):
        modifier = Modifier(location="Tokyo", deontic_modality=DeonticModality.FORBIDDEN)
        self.statement_service.create(
            user_id=self.user.id,
            agent_id=self.agent.id,
            subject_notion=self.subject_notion,
            relation=self.relation,
            complement_notion=self.complement_notion,
            modifier=modifier,
        )

        results = self.statement_service.find_by_object_and_relation(
            self.complement_notion, Relation.from_string("sell"), modifier=Modifier(location="Tokyo")
        )
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].modifier.location, "Tokyo")
        self.assertEqual(results[0].modifier.deontic_modality, DeonticModality.FORBIDDEN)


if __name__ == "__main__":
    unittest.main()


from .constraint_operator import ConstraintOperator
from .criterion import Criterion
from .relationship_type_positivity_threshold import RelationshipTypePositivityThreshold
from .relationship_type import RelationshipType
from .certainty import Certainty, certainty_from_value
from .embedding_provider import EmbeddingProvider
from .embedding_type import EmbeddingType

__all__ = [
	"ConstraintOperator",
	"Criterion",
	"RelationshipTypePositivityThreshold",
	"RelationshipType",
	"Certainty",
	"certainty_from_value",
	"EmbeddingProvider",
	"EmbeddingType",
]



import unittest
from unittest.mock import Mock
from analogai.foundation.modules.user.user import User
from analogai.foundation.infrastructure.repositories.mongodb.mongodb_user_repository_impl import MongoDBUserRepositoryImpl
from analogai.foundation.infrastructure.tests.repositories.mongodb.test_config import MongoDBEnvironmentCheckMixin

class TestMongoDBUserRepositoryImpl(MongoDBEnvironmentCheckMixin, unittest.TestCase):
    def setUp(self):
        self.mock_storage_adapter = Mock()
        self.repository = MongoDBUserRepositoryImpl(self.mock_storage_adapter)
        self.test_user = User(
            id="user_id",
            name="Test User",
            email="test@example.com",
            nationality="USA",
            location="NY"
        )
        self.collection = "users"

    def test_create_calls_store_and_returns_user(self):
        result = self.repository.create(self.test_user)
        
        self.mock_storage_adapter.store.assert_called_once()
        call_args = self.mock_storage_adapter.store.call_args
        self.assertEqual(call_args[0][0], self.collection)
        self.assertEqual(result, self.test_user)
        self.assertEqual(result.id, "user_id")
        self.assertEqual(result.email, "test@example.com")
        self.assertEqual(result.name, "Test User")

    def test_find_by_id_existing(self):
        mock_doc = {
            "id": "user_id",
            "name": "Test User",
            "email": "test@example.com",
            "nationality": "USA",
            "location": "NY",
            "timezone": "America/New_York"
        }
        self.mock_storage_adapter.retrieve_by_id.return_value = mock_doc
        
        result = self.repository.find_by_id("user_id")
        
        self.mock_storage_adapter.retrieve_by_id.assert_called_once_with(self.collection, "user_id")
        self.assertEqual(result.id, "user_id")
        self.assertEqual(result.email, "test@example.com")
        self.assertEqual(result.name, "Test User")
        self.assertEqual(result.nationality, "USA")
        self.assertEqual(result.location, "NY")
        self.assertEqual(result.timezone, "America/New_York")

    def test_find_by_id_nonexistent(self):
        self.mock_storage_adapter.retrieve_by_id.return_value = None
        
        result = self.repository.find_by_id("nonexistent_id")
        
        self.mock_storage_adapter.retrieve_by_id.assert_called_once_with(self.collection, "nonexistent_id")
        self.assertIsNone(result)

    def test_find_by_email_existing(self):
        mock_doc = {
            "id": "user_id",
            "name": "Test User",
            "email": "test@example.com",
            "nationality": "USA",
            "location": "NY",
            "timezone": "America/New_York"
        }
        self.mock_storage_adapter.retrieve_by_attributes.return_value = [mock_doc]
        
        result = self.repository.find_by_email("test@example.com")
        
        self.mock_storage_adapter.retrieve_by_attributes.assert_called_once_with(self.collection, {"email": "test@example.com"})
        self.assertEqual(result.id, "user_id")
        self.assertEqual(result.email, "test@example.com")
        self.assertEqual(result.name, "Test User")
        self.assertEqual(result.timezone, "America/New_York")

    def test_find_by_email_nonexistent(self):
        self.mock_storage_adapter.retrieve_by_attributes.return_value = []
        
        result = self.repository.find_by_email("nonexistent@example.com")
        
        self.mock_storage_adapter.retrieve_by_attributes.assert_called_once_with(self.collection, {"email": "nonexistent@example.com"})
        self.assertIsNone(result)

    def test_update_user(self):
        updated_user = User(
            id="user_id",
            name="Updated Name",
            email="updated@example.com",
            nationality="Canada",
            timezone="Europe/London"
        )
        
        result = self.repository.update(updated_user)
        
        self.mock_storage_adapter.store.assert_called_once()
        call_args = self.mock_storage_adapter.store.call_args
        self.assertEqual(call_args[0][0], self.collection)
        self.assertEqual(result, updated_user)
        self.assertEqual(result.name, "Updated Name")
        self.assertEqual(result.email, "updated@example.com")
        self.assertEqual(result.nationality, "Canada")
        self.assertEqual(result.timezone, "Europe/London")
    
    def test_user_with_timezone(self):
        timezone = "Asia/Tokyo"
        user = User(
            id="tz_user_id",
            name="Timezone User",
            email="tz@example.com",
            timezone=timezone
        )
        
        result = self.repository.create(user)
        self.assertEqual(result.timezone, timezone)
    
    def test_user_default_timezone(self):
        user = User(id="default_tz_id", name="Default TZ User", email="default_tz@example.com")
        
        self.assertIsNotNone(user.timezone)
        result = self.repository.create(user)
        self.assertIsNotNone(result.timezone)

if __name__ == '__main__':
    unittest.main()



import asyncio
import unittest
from unittest.mock import Mock, patch
from fastapi import Request
from analogai.foundation.modules.agent.agent import Agent, Role
from analogai.foundation.modules.user.user import User
from analogai.foundation.extensions.authorization.authorization_factory import AuthorizationFactory
from analogai.foundation.extensions.authorization.core.enums.operation import AuthorizationOperation
from analogai.foundation.extensions.authorization.models import AuthorizationContext

class TestAuthorizationWorkflow(unittest.TestCase):
    def setUp(self):
        self.agent_repository = Mock()
        self.user = User(id="user-123", name="Test User", email="test@example.com")
        self.agent = Agent(
            id="agent-456",
            knowledge_base="kb",
            creator=self.user,
            roles=[Role(user=self.user, authority=0.9)],
        )

    @patch("analogai.foundation.extensions.authorization.authorization_factory.create_repository_strategy")
    @patch("analogai.foundation.extensions.authorization.authorization_factory.AuthenticationFactory")
    def test_authorization_dependency_returns_context_with_role(self, authentication_factory_cls, repository_selector):
        auth_factory = authentication_factory_cls.return_value
        repository_selector.return_value.get_agent_repository.return_value = self.agent_repository

        async def auth_dependency(request: Request) -> User:
            return self.user

        auth_factory.create_authentication_dependency.return_value = auth_dependency
        self.agent_repository.find_by_id.return_value = self.agent

        factory = AuthorizationFactory()
        dependency = factory.create_authorization_dependency()

        request = Mock(spec=Request)
        request.headers = {"Authorization": "Bearer auth-token"}

        context = asyncio.run(dependency(request, agent_id="agent-456"))

        authentication_factory_cls.assert_called_once_with()
        self.agent_repository.find_by_id.assert_called_once_with("agent-456")
        self.assertIsInstance(context, AuthorizationContext)
        self.assertEqual(context.user, self.user)
        self.assertEqual(context.agent, self.agent)
        expected_authority = self.agent.roles[0].authority
        self.assertEqual(context.authority, expected_authority)
        self.assertEqual(context.operation, AuthorizationOperation.CHAT)

    @patch("analogai.foundation.extensions.authorization.authorization_factory.create_repository_strategy")
    @patch("analogai.foundation.extensions.authorization.authorization_factory.AuthenticationFactory")
    def test_authorization_dependency_returns_context_without_role(self, authentication_factory_cls, repository_selector):
        auth_factory = authentication_factory_cls.return_value
        repository_selector.return_value.get_agent_repository.return_value = self.agent_repository
        
        user_without_role = User(id="user-789", name="User Without Role", email="no-role@example.com")
        agent_without_user_role = Agent(
            id="agent-789",
            knowledge_base="kb",
            creator=self.user,
            roles=[Role(user=self.user, authority=0.9)],
        )

        async def auth_dependency(request: Request) -> User:
            return user_without_role

        auth_factory.create_authentication_dependency.return_value = auth_dependency
        self.agent_repository.find_by_id.return_value = agent_without_user_role

        factory = AuthorizationFactory()
        dependency = factory.create_authorization_dependency()

        request = Mock(spec=Request)
        request.headers = {"Authorization": "Bearer auth-token"}

        context = asyncio.run(dependency(request, agent_id="agent-789"))

        self.assertIsInstance(context, AuthorizationContext)
        self.assertEqual(context.user, user_without_role)
        self.assertEqual(context.agent, agent_without_user_role)
        self.assertEqual(context.authority, 0.0)
        self.assertEqual(context.operation, AuthorizationOperation.CHAT)

    @patch("analogai.foundation.extensions.authorization.authorization_factory.create_repository_strategy")
    @patch("analogai.foundation.extensions.authorization.authorization_factory.AuthenticationFactory")
    def test_authority_value_matches_role_authority(self, authentication_factory_cls, repository_selector):
        auth_factory = authentication_factory_cls.return_value
        repository_selector.return_value.get_agent_repository.return_value = self.agent_repository
        
        test_user = User(id="user-test", name="Test User", email="test@example.com")
        test_authority = 0.75
        test_agent = Agent(
            id="agent-test",
            knowledge_base="kb",
            creator=test_user,
            roles=[Role(user=test_user, authority=test_authority)],
        )

        async def auth_dependency(request: Request) -> User:
            return test_user

        auth_factory.create_authentication_dependency.return_value = auth_dependency
        self.agent_repository.find_by_id.return_value = test_agent

        factory = AuthorizationFactory()
        dependency = factory.create_authorization_dependency()

        request = Mock(spec=Request)
        request.headers = {"Authorization": "Bearer auth-token"}

        context = asyncio.run(dependency(request, agent_id="agent-test"))

        user_role = next(role for role in test_agent.roles if role.user.id == test_user.id)
        self.assertEqual(context.authority, user_role.authority)
        self.assertEqual(context.authority, test_authority)

if __name__ == "__main__":
    unittest.main()




from __future__ import annotations

from dataclasses import dataclass
from typing import Iterable, Protocol


class AttributeKeyResolver(Protocol):
    def can_resolve(self, attribute: object) -> bool:
        ...

    def resolve(self, attribute: object) -> object:
        ...


@dataclass(frozen=True)
class TaggedAttributeKeyResolver:
    def can_resolve(self, attribute: object) -> bool:
        return all(hasattr(attribute, name) for name in ("tag", "value", "unit"))

    def resolve(self, attribute: object) -> object:
        return (
            getattr(attribute, "tag", None),
            getattr(attribute, "value", None),
            getattr(attribute, "unit", None),
        )


@dataclass(frozen=True)
class IdentityAttributeKeyResolver:
    def can_resolve(self, attribute: object) -> bool:
        return True

    def resolve(self, attribute: object) -> object:
        return attribute


class AttributeKeyResolverChain:
    def __init__(self, resolvers: Iterable[AttributeKeyResolver] | None = None):
        self._resolvers = list(resolvers) if resolvers else [
            TaggedAttributeKeyResolver(),
            IdentityAttributeKeyResolver(),
        ]

    def resolve(self, attribute: object) -> object:
        for resolver in self._resolvers:
            if resolver.can_resolve(attribute):
                return resolver.resolve(attribute)
        return attribute



import unittest
from datetime import datetime, timedelta
from analogai.foundation.common.enums.deontic_modality import DeonticModality
from analogai.foundation.common.utils.statement_serializers import StatementSerializer, serialize_statement
from analogai.foundation.core.value_objects.modifier import Modifier
from analogai.foundation.core.value_objects.relation import Relation
from analogai.foundation.core.value_objects.time import Time


class TestStatementSerializer(unittest.TestCase):
    
    def test_serialize_basic(self):
        serializer = StatementSerializer.create_default()
        result = serializer.serialize("Human", "is", "mortal", 0.8)
        self.assertIn("Human", result)
        self.assertIn("mortal", result)
    
    def test_serialize_negative_probability(self):
        serializer = StatementSerializer.create_default()
        result = serializer.serialize("Human", "is", "mortal", 0.2)
        self.assertIn("not", result.lower())
    
    def test_serialize_uncertain_probability(self):
        serializer = StatementSerializer.create_default()
        result = serializer.serialize("Human", "is", "mortal", 0.5)
        self.assertIn("might", result.lower())
    
    def test_serialize_statement_with_modifier(self):
        today = datetime.now()
        modifier = Modifier(
            location="Rome",
            from_time=Time(year=today.year, month=today.month, day=today.day)
        )
        result = serialize_statement(
            "Human",
            "mortal",
            Relation.from_string("is"),
            0.8,
            modifier
        )
        self.assertIn("human", result.lower())
        self.assertIn("mortal", result.lower())
        self.assertIn("rome", result.lower())
    
    def test_serialize_statement_with_timezone(self):
        today = datetime.now()
        modifier = Modifier(
            location="New York",
            from_time=Time(year=today.year, month=today.month, day=today.day)
        )
        result = serialize_statement(
            "Human",
            "mortal",
            Relation.from_string("is"),
            0.8,
            modifier,
            timezone="America/New_York"
        )
        self.assertIn("human", result.lower())
        self.assertIn("mortal", result.lower())
        self.assertIn("new york", result.lower())
    
    def test_serialize_statement_with_different_timezones(self):
        today = datetime.now()
        modifier = Modifier(
            from_time=Time(year=today.year, month=today.month, day=today.day)
        )
        
        result_ny = serialize_statement(
            "Human",
            "mortal",
            Relation.from_string("is"),
            0.8,
            modifier,
            timezone="America/New_York"
        )
        
        result_london = serialize_statement(
            "Human",
            "mortal",
            Relation.from_string("is"),
            0.8,
            modifier,
            timezone="Europe/London"
        )
        
        result_tokyo = serialize_statement(
            "Human",
            "mortal",
            Relation.from_string("is"),
            0.8,
            modifier,
            timezone="Asia/Tokyo"
        )
        
        self.assertIn("human", result_ny.lower())
        self.assertIn("mortal", result_ny.lower())
        self.assertIn("human", result_london.lower())
        self.assertIn("mortal", result_london.lower())
        self.assertIn("human", result_tokyo.lower())
        self.assertIn("mortal", result_tokyo.lower())
    
    def test_serialize_statement_without_timezone(self):
        today = datetime.now()
        modifier = Modifier(
            from_time=Time(year=today.year, month=today.month, day=today.day)
        )
        result = serialize_statement(
            "Human",
            "mortal",
            Relation.from_string("is"),
            0.8,
            modifier
        )
        self.assertIn("human", result.lower())
        self.assertIn("mortal", result.lower())
    
    def test_serialize_statement_with_obligatory_deontic_modality(self):
        modifier = Modifier(deontic_modality=DeonticModality.OBLIGATORY)
        result = serialize_statement(
            "aristotle",
            "iphone",
            Relation.from_string("do"),
            0.8,
            modifier
        )
        self.assertIn("aristotle", result.lower())
        self.assertIn("iphone", result.lower())
        self.assertIn("must", result.lower())
        self.assertNotIn("it is obligatory", result.lower())
    
    def test_serialize_statement_with_permitted_deontic_modality(self):
        modifier = Modifier(deontic_modality=DeonticModality.PERMITTED)
        result = serialize_statement(
            "aristotle",
            "iphone",
            Relation.from_string("do"),
            0.8,
            modifier
        )
        self.assertIn("aristotle", result.lower())
        self.assertIn("iphone", result.lower())
        self.assertIn("can", result.lower())
    
    def test_serialize_statement_with_forbidden_deontic_modality(self):
        modifier = Modifier(deontic_modality=DeonticModality.FORBIDDEN)
        result = serialize_statement(
            "aristotle",
            "iphone",
            Relation.from_string("do"),
            0.8,
            modifier
        )
        self.assertIn("aristotle", result.lower())
        self.assertIn("iphone", result.lower())
        self.assertIn("can not", result.lower())


if __name__ == '__main__':
    unittest.main()



from datetime import datetime
from typing import Dict, Any, Optional, Tuple, Union, List

from analogai.foundation.modules.belief.belief import Belief
from analogai.foundation.common.enums.relationship_type import RelationshipType
from analogai.foundation.core.value_objects.relation import Relation
from analogai.foundation.common.enums.criterion import Criterion
from analogai.foundation.common.enums.constraint_operator import ConstraintOperator
from analogai.foundation.common.utils.time_utils.time_converter import TimeConverter
from analogai.foundation.core.value_objects.time import Time
from analogai.foundation.common.utils.similarity.similarity_service import SimilarityService
from analogai.foundation.settings import config
from analogai.foundation.common.logging.app_logger import app_logger


def _environment() -> str:
    value = config.get("ENVIRONMENT")
    return str(value).strip().lower()


def _use_vector_search() -> bool:
    if _environment() in {"test", "testing"}:
        return False
    return bool(config.get("vector.enabled", True))


def _similarity_cutoff() -> float:
    value = config.get("vector.cutoff", 0.4)
    try:
        return float(value)
    except (TypeError, ValueError):
        return 0.4


class Condition:
    def __init__(
        self,
        criterion: Criterion,
        operator: ConstraintOperator,
        value: Optional[Any] = None,
    ):
        self.criterion = criterion
        self.operator = operator
        self.value = value


CriterionConstraint = Union[Any, Tuple[Union[str, ConstraintOperator], Any], Condition]


class BeliefCriteriaMatcher:
    @staticmethod
    def matches_clause(
        belief: Belief,
        clause: Dict[Criterion, Union[CriterionConstraint, List[CriterionConstraint]]],
        similarity_service: Optional[SimilarityService] = None,
    ) -> bool:
        for criterion, constraint in clause.items():
            if not BeliefCriteriaMatcher._criterion_matches_constraint(
                belief, criterion, constraint, similarity_service
            ):
                return False
        return True

    @staticmethod
    def _criterion_matches_constraint(
        belief: Belief,
        criterion: Criterion,
        constraint: Union[CriterionConstraint, List[CriterionConstraint]],
        similarity_service: Optional[SimilarityService] = None,
    ) -> bool:
        if type(constraint) is list:
            if constraint and (type(constraint[0]) is tuple or type(constraint[0]) is list) and len(constraint[0]) >= 1:
                first = constraint[0]
                if BeliefCriteriaMatcher._is_operator_constraint(first[0]):
                    return all(
                        BeliefCriteriaMatcher._criterion_matches_constraint(
                            belief, criterion, c, similarity_service
                        )
                        for c in constraint
                    )
            return any(
                BeliefCriteriaMatcher._criterion_matches_from_condition(
                    belief, Condition(criterion, ConstraintOperator.Eq, v), similarity_service
                )
                for v in constraint
            )
        if type(constraint) is Condition:
            return BeliefCriteriaMatcher._criterion_matches_from_condition(
                belief, constraint, similarity_service
            )
        if (type(constraint) is tuple or type(constraint) is list) and len(constraint) >= 1:
            operator = BeliefCriteriaMatcher._normalize_operator(constraint[0])
            value = constraint[1] if len(constraint) > 1 else None
            return BeliefCriteriaMatcher._criterion_matches_from_condition(
                belief, Condition(criterion, operator, value), similarity_service
            )
        return BeliefCriteriaMatcher._criterion_matches_from_condition(
            belief, Condition(criterion, ConstraintOperator.Eq, constraint), similarity_service
        )

    @staticmethod
    def _is_operator_constraint(first: Any) -> bool:
        if type(first) is ConstraintOperator:
            return first in (
                ConstraintOperator.Min,
                ConstraintOperator.Max,
                ConstraintOperator.NotNull,
                ConstraintOperator.Null,
            )
        if type(first) is str:
            return first in ("min", "max", "not_null", "null")
        return False

    @staticmethod
    def _normalize_operator(op: Union[str, ConstraintOperator]) -> ConstraintOperator:
        if type(op) is ConstraintOperator:
            return op
        return ConstraintOperator(op)

    @staticmethod
    def _criterion_matches_from_condition(
        belief: Belief,
        condition: Condition,
        similarity_service: Optional[SimilarityService] = None,
    ) -> bool:
        operator = condition.operator
        value = condition.value
        if operator == ConstraintOperator.Eq:
            return BeliefCriteriaMatcher._criterion_equals(
                belief, condition.criterion, value, similarity_service
            )
        if operator == ConstraintOperator.Min:
            return BeliefCriteriaMatcher._criterion_min(belief, condition.criterion, value)
        if operator == ConstraintOperator.Max:
            return BeliefCriteriaMatcher._criterion_max(belief, condition.criterion, value)
        if operator == ConstraintOperator.NotNull:
            return BeliefCriteriaMatcher._criterion_not_null(belief, condition.criterion)
        if operator == ConstraintOperator.Null:
            return BeliefCriteriaMatcher._criterion_null(belief, condition.criterion)
        return True

    @staticmethod
    def _criterion_equals(
        belief: Belief,
        criterion: Criterion,
        expected: Any,
        similarity_service: Optional[SimilarityService] = None,
    ) -> bool:
        if criterion == Criterion.SubjectCaption:
            return BeliefCriteriaMatcher._text_matches(
                belief.subject_notion.caption,
                expected,
                similarity_service,
            )
        if criterion in (Criterion.ObjectCaption, Criterion.ComplementCaption):
            return BeliefCriteriaMatcher._text_matches(
                belief.complement_notion.caption,
                expected,
                similarity_service,
            )
        if criterion == Criterion.Relation:
            if expected is None:
                return False
            if isinstance(expected, (Relation, RelationshipType)):
                if belief.relation == expected:
                    return True
                if similarity_service and _use_vector_search():
                    score = similarity_service.similarity(
                        belief.relation.value,
                        expected.value,
                    )
                    threshold = _similarity_cutoff()
                    app_logger.debug(
                        "BeliefQuery relation similarity: actual='%s' expected='%s' score=%.4f threshold=%.4f",
                        belief.relation.value,
                        expected.value,
                        score,
                        threshold,
                    )
                    return score >= threshold
                return False
            raise TypeError("Criterion.Relation expects a Relation or RelationshipType")
        if criterion == Criterion.Location:
            return BeliefCriteriaMatcher._belief_has_location(belief, expected)
        if criterion == Criterion.DeonticModality:
            return BeliefCriteriaMatcher._belief_deontic_modality_equals(belief, expected)
        if criterion == Criterion.AttributeTag:
            return BeliefCriteriaMatcher._belief_has_measure_kind(belief, expected)
        if criterion == Criterion.AttributeUnit:
            return BeliefCriteriaMatcher._belief_has_measure_unit(belief, expected)
        if criterion == Criterion.AttributeValue:
            return BeliefCriteriaMatcher._belief_has_measure_value(belief, expected)
        return True

    @staticmethod
    def _criterion_min(belief: Belief, criterion: Criterion, expected: Any) -> bool:
        if criterion == Criterion.Quantity:
            quantity = belief.modifier.quantity if belief.modifier else None
            return quantity is not None and expected is not None and quantity >= float(expected)
        if criterion == Criterion.Time:
            return BeliefCriteriaMatcher._belief_has_time_ge(belief, expected)
        if criterion == Criterion.AttributeMinValue:
            return BeliefCriteriaMatcher._belief_has_measure_number_ge(belief, expected)
        if criterion == Criterion.Probability:
            return expected is not None and float(belief.probability) >= float(expected)
        return True

    @staticmethod
    def _criterion_max(belief: Belief, criterion: Criterion, expected: Any) -> bool:
        if criterion == Criterion.Quantity:
            quantity = belief.modifier.quantity if belief.modifier else None
            return quantity is not None and expected is not None and quantity <= float(expected)
        if criterion == Criterion.Time:
            return BeliefCriteriaMatcher._belief_has_time_le(belief, expected)
        if criterion == Criterion.AttributeMaxValue:
            return BeliefCriteriaMatcher._belief_has_measure_number_le(belief, expected)
        if criterion == Criterion.Probability:
            return expected is not None and float(belief.probability) <= float(expected)
        return True

    @staticmethod
    def _criterion_not_null(belief: Belief, criterion: Criterion) -> bool:
        if criterion == Criterion.Time:
            return BeliefCriteriaMatcher._belief_has_time(belief)
        if criterion == Criterion.DeonticModality:
            return BeliefCriteriaMatcher._belief_has_deontic_modality(belief)
        if criterion in (Criterion.ObjectCaption, Criterion.ComplementCaption):
            return belief.complement_notion.caption != ""
        if criterion == Criterion.Quantity:
            return belief.modifier is not None and belief.modifier.quantity is not None
        if criterion == Criterion.AttributeTag:
            return BeliefCriteriaMatcher._belief_has_any_measure(belief)
        return True

    @staticmethod
    def _criterion_null(belief: Belief, criterion: Criterion) -> bool:
        if criterion == Criterion.Time:
            return not BeliefCriteriaMatcher._belief_has_time(belief)
        if criterion == Criterion.DeonticModality:
            return not BeliefCriteriaMatcher._belief_has_deontic_modality(belief)
        if criterion in (Criterion.ObjectCaption, Criterion.ComplementCaption):
            return belief.complement_notion.caption == ""
        if criterion == Criterion.Quantity:
            return belief.modifier is None or belief.modifier.quantity is None
        if criterion == Criterion.AttributeTag:
            return not BeliefCriteriaMatcher._belief_has_any_measure(belief)
        return False

    @staticmethod
    def _belief_has_any_measure(belief: Belief) -> bool:
        measures = BeliefCriteriaMatcher._collect_notion_measures(belief)
        return len(measures) > 0

    @staticmethod
    def _belief_has_measure_kind(belief: Belief, expected: Any) -> bool:
        if expected is None:
            return False
        expected_kind = str(expected).lower()
        measures = BeliefCriteriaMatcher._collect_notion_measures(belief)
        return any(m.tag.lower() == expected_kind for m in measures)

    @staticmethod
    def _belief_has_measure_unit(belief: Belief, expected: Any) -> bool:
        if expected is None:
            return False
        expected_unit = str(expected).lower()
        measures = BeliefCriteriaMatcher._collect_notion_measures(belief)
        return any(m.unit is not None and m.unit.lower() == expected_unit for m in measures)

    @staticmethod
    def _belief_has_measure_value(belief: Belief, expected: Any) -> bool:
        if expected is None:
            return False
        measures = BeliefCriteriaMatcher._collect_notion_measures(belief)
        if type(expected) is tuple and len(expected) >= 2:
            kind = str(expected[0]).lower()
            value = expected[1]
            unit = expected[2] if len(expected) > 2 else None
            for m in measures:
                if m.tag.lower() != kind:
                    continue
                if unit is not None and (m.unit is None or m.unit.lower() != str(unit).lower()):
                    continue
                return m.value == value
            return False
        return any(m.value == expected for m in measures)

    @staticmethod
    def _belief_has_measure_number_ge(belief: Belief, expected: Any) -> bool:
        kind, unit, number = BeliefCriteriaMatcher._parse_measure_constraint(expected)
        if kind is None or number is None:
            return False
        return BeliefCriteriaMatcher._belief_has_measure_number_compare(
            belief, kind, unit, number, ConstraintOperator.Ge
        )

    @staticmethod
    def _belief_has_measure_number_le(belief: Belief, expected: Any) -> bool:
        kind, unit, number = BeliefCriteriaMatcher._parse_measure_constraint(expected)
        if kind is None or number is None:
            return False
        return BeliefCriteriaMatcher._belief_has_measure_number_compare(
            belief, kind, unit, number, ConstraintOperator.Le
        )

    @staticmethod
    def _parse_measure_constraint(expected: Any) -> tuple[Optional[str], Optional[str], Optional[float]]:
        if type(expected) is tuple and len(expected) >= 2:
            kind = str(expected[0]).lower()
            number = float(expected[1]) if expected[1] is not None else None
            unit = str(expected[2]).lower() if len(expected) > 2 and expected[2] is not None else None
            return kind, unit, number
        return None, None, None

    @staticmethod
    def _belief_has_measure_number_compare(
        belief: Belief,
        kind: str,
        unit: Optional[str],
        number: float,
        op: ConstraintOperator,
    ) -> bool:
        measures = BeliefCriteriaMatcher._collect_notion_measures(belief)
        if not measures:
            return False
        for m in measures:
            if m.tag.lower() != kind:
                continue
            if unit is not None and (m.unit is None or m.unit.lower() != unit):
                continue
            try:
                actual = float(m.value)
            except (TypeError, ValueError):
                continue
            if op == ConstraintOperator.Ge and actual >= number:
                return True
            if op == ConstraintOperator.Le and actual <= number:
                return True
        return False

    @staticmethod
    def _belief_has_time(belief: Belief) -> bool:
        modifier = belief.modifier
        if modifier is None:
            return False
        return TimeConverter.has_time_data(modifier.from_time) or TimeConverter.has_time_data(modifier.to_time)

    @staticmethod
    def _belief_has_time_ge(belief: Belief, expected: Any) -> bool:
        learning_time = BeliefCriteriaMatcher._get_belief_earliest_time(belief)
        if learning_time is None or expected is None:
            return False
        return BeliefCriteriaMatcher._time_compare_ge(learning_time, expected)

    @staticmethod
    def _belief_has_time_le(belief: Belief, expected: Any) -> bool:
        learning_time = BeliefCriteriaMatcher._get_belief_latest_time(belief)
        if learning_time is None or expected is None:
            return False
        return BeliefCriteriaMatcher._time_compare_le(learning_time, expected)

    @staticmethod
    def _belief_has_location(belief: Belief, expected: Any) -> bool:
        modifier = belief.modifier
        if modifier is None or expected is None:
            return False
        location = modifier.location
        return location is not None and str(location).lower() == str(expected).lower()

    @staticmethod
    def _belief_has_deontic_modality(belief: Belief) -> bool:
        modifier = belief.modifier
        if modifier is None:
            return False
        return modifier.deontic_modality is not None

    @staticmethod
    def _belief_deontic_modality_equals(belief: Belief, expected: Any) -> bool:
        modifier = belief.modifier
        if modifier is None or expected is None:
            return False
        actual = modifier.deontic_modality
        if actual is None:
            return False
        return BeliefCriteriaMatcher._deontic_modality_matches(actual, expected)

    @staticmethod
    def _deontic_modality_matches(actual: Any, expected: Any) -> bool:
        if actual == expected:
            return True
        try:
            actual_str = actual.value
        except AttributeError:
            actual_str = str(actual)
        try:
            expected_str = expected.value
        except AttributeError:
            expected_str = str(expected)
        return str(actual_str).lower() == str(expected_str).lower()

    @staticmethod
    def _get_belief_earliest_time(belief: Belief) -> Optional[Time]:
        modifier = belief.modifier
        if modifier is None:
            return None
        from_time = modifier.from_time
        to_time = modifier.to_time
        if TimeConverter.has_time_data(from_time):
            return from_time
        if TimeConverter.has_time_data(to_time):
            return to_time
        return None

    @staticmethod
    def _get_belief_latest_time(belief: Belief) -> Optional[Time]:
        modifier = belief.modifier
        if modifier is None:
            return None
        from_time = modifier.from_time
        to_time = modifier.to_time
        if TimeConverter.has_time_data(to_time):
            return to_time
        if TimeConverter.has_time_data(from_time):
            return from_time
        return None

    @staticmethod
    def _time_to_datetime(time_or_dt: Any) -> Optional[datetime]:
        if time_or_dt is None:
            return None
        if type(time_or_dt) is datetime:
            return time_or_dt
        if type(time_or_dt) is Time:
            reference = datetime.now().astimezone()
            return TimeConverter.to_datetime(time_or_dt, reference)
        return None

    @staticmethod
    def _time_compare_ge(learning_time: Time, expected: Any) -> bool:
        learning_dt = BeliefCriteriaMatcher._time_to_datetime(learning_time)
        expected_dt = BeliefCriteriaMatcher._time_to_datetime(expected)
        if learning_dt is None or expected_dt is None:
            return False
        return learning_dt >= expected_dt

    @staticmethod
    def _time_compare_le(learning_time: Time, expected: Any) -> bool:
        learning_dt = BeliefCriteriaMatcher._time_to_datetime(learning_time)
        expected_dt = BeliefCriteriaMatcher._time_to_datetime(expected)
        if learning_dt is None or expected_dt is None:
            return False
        return learning_dt <= expected_dt

    @staticmethod
    def _collect_notion_measures(belief: Belief) -> List:
        measures: List = []
        if belief.subject_notion and belief.subject_notion.attributes:
            measures.extend(belief.subject_notion.attributes)
        if belief.complement_notion and belief.complement_notion.attributes:
            measures.extend(belief.complement_notion.attributes)
        return measures

    @staticmethod
    def _text_matches(
        actual: str,
        expected: Any,
        similarity_service: Optional[SimilarityService] = None,
    ) -> bool:
        if expected is None:
            return False
        expected_str = str(expected)
        if actual.lower() == expected_str.lower():
            return True
        if _use_vector_search():
            raise RuntimeError(
                "Caption similarity fallback disabled while USE_VECTOR_SEARCH is enabled"
            )
        return False



from __future__ import annotations

from pathlib import Path
from typing import List, Optional, TYPE_CHECKING
import os

from analogai.foundation.common.logging.app_logger import app_logger

if TYPE_CHECKING:
    from sentence_transformers import SentenceTransformer


class LocalSentenceTransformerProvider:
    def __init__(self, model_name: str = "all-MiniLM-L6-v2", cache_folder: Optional[str] = None):
        self.model_name = model_name
        self._model: Optional["SentenceTransformer"] = None
        self._dimension: Optional[int] = None
        self.cache_folder = cache_folder or os.getenv(
            "SENTENCE_TRANSFORMERS_CACHE",
            str(Path.home() / ".cache" / "sentence_transformers"),
        )
        Path(self.cache_folder).mkdir(parents=True, exist_ok=True)
        app_logger.info(
            f"Local embeddings initializing with model: {model_name}, cache: {self.cache_folder}"
        )
        from sentence_transformers import SentenceTransformer

        app_logger.info(f"Loading model {self.model_name} from cache: {self.cache_folder}")
        self._model = SentenceTransformer(self.model_name, cache_folder=self.cache_folder)
        self._dimension = self._model.get_sentence_embedding_dimension()
        app_logger.info(f"Local embedding model loaded with dimension: {self._dimension}")

    @property
    def model(self) -> "SentenceTransformer":
        if self._model is None:
            raise RuntimeError(
                "Local embedding model not initialized. "
                "Instantiate only when USE_VECTOR_SEARCH is enabled."
            )
        return self._model

    @property
    def dimension(self) -> int | None:
        if self._dimension is None:
            if self._model is None:
                return None
            self._dimension = self._model.get_sentence_embedding_dimension()
        return self._dimension

    def encode(self, text: str) -> List[float]:
        if not text or not text.strip():
            if self.dimension is None:
                return []
            app_logger.warning("Attempted to encode empty text, returning zero vector")
            return [0.0] * self.dimension
        return self.model.encode(text, convert_to_numpy=True).tolist()

    def encode_batch(self, texts: List[str]) -> List[List[float]]:
        if not texts:
            return []
        filtered_texts = [text if text and text.strip() else "" for text in texts]
        return self.model.encode(filtered_texts, convert_to_numpy=True).tolist()



from typing import Optional, List, Dict
from analogai.foundation.extensions.fallback.repositories.fallback_repository import FallbackRepository
from analogai.foundation.extensions.fallback import Fallback
from analogai.foundation.infrastructure.storage.in_memory import InMemoryStorageAdapterImpl
from analogai.foundation.extensions.fallback.mappers.fallback_storage_mapper import FallbackStorageMapper

class InMemoryFallbackRepositoryImpl(FallbackRepository):
    COLLECTION_NAME = "fallbacks"

    def __init__(self):
        self.storage_adapter = InMemoryStorageAdapterImpl()
        self._fallback_cache: Dict[str, Fallback] = {}

    def create(self, fallback: Fallback) -> Fallback:
        data = FallbackStorageMapper.to_dict(fallback)
        stored_data = self.storage_adapter.store(self.COLLECTION_NAME, data)
        if not stored_data:
            raise ValueError(f"Failed to store fallback {fallback.id}")
        self._fallback_cache[fallback.id] = fallback
        return fallback

    def find_by_id(self, fallback_id: str) -> Optional[Fallback]:
        if fallback_id in self._fallback_cache:
            return self._fallback_cache[fallback_id]
        data = self.storage_adapter.retrieve_by_id(self.COLLECTION_NAME, fallback_id)
        if not data:
            return None
        fallback = FallbackStorageMapper.from_dict(data)
        self._fallback_cache[fallback_id] = fallback
        return fallback

    def find_by_agent_id(self, agent_id: str) -> List[Fallback]:
        fallbacks_data = self.storage_adapter.retrieve_by_attributes(
            self.COLLECTION_NAME,
            {"agent_id": agent_id}
        )
        fallbacks = []
        for data in fallbacks_data:
            fallback_id = data.get("id")
            if fallback_id in self._fallback_cache:
                fallbacks.append(self._fallback_cache[fallback_id])
            else:
                fallback = FallbackStorageMapper.from_dict(data)
                self._fallback_cache[fallback_id] = fallback
                fallbacks.append(fallback)
        return fallbacks



from unittest import TestCase
import unittest

from analogai.foundation.common.dtos.notion_expression import NotionExpression
from analogai.foundation.core.value_objects.relation import Relation
from analogai.foundation.core.value_objects.cause import Cause
from analogai.foundation.core.value_objects.effect import Effect
from analogai.foundation.common.enums.criterion import Criterion
from analogai.foundation.common.enums.certainty import Certainty, certainty_from_value
from analogai.foundation.tests.functional.services.setup import ServicesTestConfig


CRITERION_SUBJECT = Criterion.SubjectCaption
CRITERION_OBJECT = Criterion.ObjectCaption
CRITERION_RELATION = Criterion.Relation
CRITERION_LOCATION = Criterion.Location
CRITERION_DEONTIC = Criterion.DeonticModality
CRITERION_FROM_TIME = Criterion.FromTime


class TestHypotheticalStatementServiceFunctional(TestCase):
    def setUp(self):
        self.config = ServicesTestConfig()
        self.user = self.config.user
        self.agent = self.config.agent
        self.user_id = self.user.id
        self.agent_id = self.agent.id

        self.notion_service = self.config.notion_service
        self.belief_service = self.config.belief_service
        self.hypothetical_statement_service = self.config.hypothetical_statement_service

    def tearDown(self):
        self.config.tear_down()

    def test_create_does_not_create_effect_belief(self):
        causes = [
            Cause({
                CRITERION_SUBJECT: "rain",
                CRITERION_OBJECT: "today",
                CRITERION_RELATION: Relation.from_string("is"),
            })
        ]
        effect = Effect({
            CRITERION_SUBJECT: "ground",
            CRITERION_OBJECT: "wet",
            CRITERION_RELATION: Relation.from_string("is"),
            Criterion.Certainty: Certainty.UNCERTAIN,
        })

        hypothetical = self.hypothetical_statement_service.create(
            user_id=self.user_id,
            agent_id=self.agent_id,
            causes=causes,
            effect=effect,
        )

        ground = self.notion_service.find_or_create(
            self.user_id,
            self.agent_id,
            NotionExpression(caption="ground"),
        )
        wet = self.notion_service.find_or_create(
            self.user_id,
            self.agent_id,
            NotionExpression(caption="wet"),
        )
        effect_belief = self.belief_service.find(
            ground,
            wet,
            Relation.from_string("is"),
        )

        self.assertIsNotNone(hypothetical)
        self.assertTrue(hypothetical.effect)
        self.assertIsNone(effect_belief)
        self.assertEqual(len(self.hypothetical_statement_service.find(self.agent_id)), 1)

    def test_create_stores_hypothetical_statement(self):
        causes = [
            Cause({
                CRITERION_SUBJECT: "rain",
                CRITERION_OBJECT: "today",
                CRITERION_RELATION: Relation.from_string("is"),
            })
        ]
        effect = Effect({
            CRITERION_SUBJECT: "ground",
            CRITERION_OBJECT: "wet",
            CRITERION_RELATION: Relation.from_string("is"),
            Criterion.Certainty: Certainty.UNCERTAIN,
        })

        hypothetical = self.hypothetical_statement_service.create(
            user_id=self.user_id,
            agent_id=self.agent_id,
            causes=causes,
            effect=effect,
        )

        stored = self.hypothetical_statement_service.find(self.agent_id)

        self.assertEqual(len(stored), 1)
        self.assertEqual(stored[0].id, hypothetical.id)
        self.assertEqual(stored[0].probability, 0.5)
        stored_certainty = certainty_from_value(
            stored[0].effect.criteria.get(Criterion.Certainty)
        )
        self.assertEqual(stored_certainty, Certainty.UNCERTAIN)

    def test_create_stores_hypothetical_statement_validity(self):
        causes = [
            Cause({
                CRITERION_SUBJECT: "rain",
                CRITERION_OBJECT: "today",
                CRITERION_RELATION: Relation.from_string("is"),
            })
        ]
        effect = Effect({
            CRITERION_SUBJECT: "ground",
            CRITERION_OBJECT: "wet",
            CRITERION_RELATION: Relation.from_string("is"),
            Criterion.Certainty: Certainty.CERTAIN,
        })

        hypothetical = self.hypothetical_statement_service.create(
            user_id=self.user_id,
            agent_id=self.agent_id,
            causes=causes,
            effect=effect,
            validity=0.55,
        )

        stored = self.hypothetical_statement_service.find(self.agent_id)

        self.assertEqual(len(stored), 1)
        self.assertEqual(stored[0].id, hypothetical.id)
        self.assertEqual(stored[0].validity, 0.55)

    def test_create_requires_effect_criteria(self):
        causes = [
            Cause({
                CRITERION_SUBJECT: "rain",
                CRITERION_OBJECT: "today",
                CRITERION_RELATION: Relation.from_string("is"),
            })
        ]
        effect = Effect({
            CRITERION_SUBJECT: "ground",
        })

        with self.assertRaises(ValueError):
            self.hypothetical_statement_service.create(
                user_id=self.user_id,
                agent_id=self.agent_id,
                causes=causes,
                effect=effect,
            )

    def test_create_preserves_effect_modifiers(self):
        causes = [
            Cause({
                CRITERION_SUBJECT: "winter",
                CRITERION_OBJECT: "paris",
                CRITERION_RELATION: Relation.from_string("in"),
            })
        ]

        effect = Effect({
            CRITERION_SUBJECT: "winter",
            CRITERION_OBJECT: "rome",
            CRITERION_RELATION: Relation.from_string("in"),
            CRITERION_LOCATION: "paris",
        })

        hypothetical = self.hypothetical_statement_service.create(
            user_id=self.user_id,
            agent_id=self.agent_id,
            causes=causes,
            effect=effect,
        )

        self.assertEqual(hypothetical.effect.criteria[CRITERION_LOCATION], "paris")

    def test_create_preserves_multiple_effect_attributes(self):
        causes = [
            Cause({
                CRITERION_SUBJECT: "exam",
                CRITERION_OBJECT: "tomorrow",
                CRITERION_RELATION: Relation.from_string("is"),
            })
        ]

        effect = Effect({
            CRITERION_SUBJECT: "student",
            CRITERION_OBJECT: "study",
            CRITERION_RELATION: Relation.from_string("must"),
            CRITERION_LOCATION: "library",
            CRITERION_DEONTIC: "obligatory",
            CRITERION_FROM_TIME: {"year": 2026, "month": 2},
        })

        hypothetical = self.hypothetical_statement_service.create(
            user_id=self.user_id,
            agent_id=self.agent_id,
            causes=causes,
            effect=effect,
        )

        self.assertEqual(hypothetical.effect.criteria[CRITERION_LOCATION], "library")
        self.assertEqual(hypothetical.effect.criteria[CRITERION_DEONTIC], "obligatory")
        self.assertEqual(
            hypothetical.effect.criteria[CRITERION_FROM_TIME],
            {"year": 2026, "month": 2},
        )


if __name__ == "__main__":
    unittest.main()



from typing import Dict, List, Optional
from datetime import datetime
from analogai.foundation.modules.belief.belief import Belief
from analogai.foundation.modules.shared.statement_base import StatementBase
from analogai.foundation.modules.notion.notion import Notion
from analogai.foundation.core.value_objects.modifier import Modifier
from analogai.foundation.core.value_objects.relation import Relation
from analogai.foundation.common.utils.modifier_utils import ModifierBuilder
from analogai.foundation.common.enums.deontic_modality import DeonticModality
from .notion_storage_mapper import NotionStorageMapper
from .time_storage_mapper import TimeStorageMapper


def _relation_to_storage(relation: Relation) -> str:
    return relation.value


def _relation_from_storage(value) -> Relation:
    if value is None or value == "":
        return Relation.from_string("")
    return Relation.from_string(str(value))


class BeliefStorageMapper:
    @staticmethod
    def to_dict(belief: Belief) -> Dict:
        result = {
            "id": belief.id,
            "agent_id": belief.agent_id,
            "subject_notion": NotionStorageMapper.to_dict(belief.subject_notion),
            "complement_notion": NotionStorageMapper.to_dict(belief.complement_notion),
            "relation": _relation_to_storage(belief.relation),
            "statement_ids": [s.id for s in belief.unfiltered_statements],
            "created_at": belief.created_at.isoformat(),
            "updated_at": belief.updated_at.isoformat()
        }
        if belief.modifier is not None and belief.modifier.quantity is not None:
            result["quantity"] = belief.modifier.quantity
        if belief.location is not None:
            result["location"] = belief.location
        if belief.from_time is not None:
            result.update(TimeStorageMapper.flatten_to_dict(belief.from_time, "from_time"))
        if belief.to_time is not None:
            result.update(TimeStorageMapper.flatten_to_dict(belief.to_time, "to_time"))
        if belief.deontic_modality is not None:
            result["deontic_modality"] = belief.deontic_modality.value
        if belief.modifier and belief.modifier.attributes is not None:
            result["attributes"] = [
                {"tag": attribute.tag, "value": attribute.value, "unit": attribute.unit}
                for attribute in belief.modifier.attributes
            ]
        return result

    @staticmethod
    def from_dict(data: Dict, statements: Optional[List[StatementBase]] = None) -> Belief:
        created_at = datetime.fromisoformat(data["created_at"]) if data.get("created_at") else datetime.now()
        updated_at = datetime.fromisoformat(data["updated_at"]) if data.get("updated_at") else datetime.now()

        belief = Belief(
            id=data.get("id"),
            agent_id=data.get("agent_id", ""),
            created_at=created_at,
            updated_at=updated_at,
        )
        stored_updated_at = updated_at
        statements_list = statements or []
        if not statements_list:
            statement_ids = data.get("statement_ids", [])
            if statement_ids:
                raise ValueError(f"Belief {data.get('id')} has statement_ids {statement_ids} but no statements were provided")
        
        modifier = None
        from_time = TimeStorageMapper.unflatten_from_dict(data, "from_time")
        if not from_time and data.get("from_time"):
            from_time = TimeStorageMapper.from_dict(data.get("from_time"))
        to_time = TimeStorageMapper.unflatten_from_dict(data, "to_time")
        if not to_time and data.get("to_time"):
            to_time = TimeStorageMapper.from_dict(data.get("to_time"))
        location = data.get("location")
        deontic_modality = data.get("deontic_modality")
        
        attributes_payload = data.get("attributes")
        if attributes_payload is None:
            attributes_payload = data.get("measures")
        if any([location, from_time, to_time, deontic_modality]) or attributes_payload is not None or data.get("quantity") is not None:
            builder = ModifierBuilder()
            if location:
                builder.with_location(location)
            if from_time:
                builder.with_from_time(
                    year=from_time.year,
                    month=from_time.month,
                    day=from_time.day,
                    hour=from_time.hour,
                    minute=from_time.minute
                )
            if to_time:
                builder.with_to_time(
                    year=to_time.year,
                    month=to_time.month,
                    day=to_time.day,
                    hour=to_time.hour,
                    minute=to_time.minute
                )
            if deontic_modality:
                builder.with_deontic_modality(DeonticModality(deontic_modality))
            modifier = builder.build()
            if attributes_payload is not None:
                from analogai.foundation.core.value_objects.attribute import Attribute
                modifier.attributes = [
                    Attribute(tag=m.get("tag") or m.get("kind"), value=m.get("value"), unit=m.get("unit"))
                    for m in attributes_payload or []
                ]
            if data.get("quantity") is not None:
                modifier.quantity = data.get("quantity")
            belief._modifier = modifier
        
        # Add statements - validation will happen in add_statement if parameters don't match
        for statement in statements_list:
            belief.add_statement(statement)
        if not belief.statements:
            raise ValueError(f"Belief {belief.id} must have at least one statement")
        belief.updated_at = stored_updated_at
        return belief




from typing import Dict, List, Optional, Any, TYPE_CHECKING
from datetime import datetime, timedelta

from analogai.foundation.modules.notion.notion_service import NotionService
from analogai.foundation.modules.direct_statement.statement_service_base import StatementServiceBase

if TYPE_CHECKING:
    from analogai.foundation.common.enums.criterion import Criterion
from analogai.foundation.modules.belief.belief import Belief
from analogai.foundation.modules.notion.notion import Notion
from analogai.foundation.modules.direct_statement.direct_statement import DirectStatement
from analogai.foundation.modules.belief.belief_assembler import BeliefAssembler
from analogai.foundation.modules.belief.belief_deduction_applier import BeliefDeductionApplier
from analogai.foundation.modules.belief.belief_filterer import BeliefFilterer, BeliefKey
from analogai.foundation.core.value_objects.modifier import Modifier
from analogai.foundation.core.value_objects.relation import Relation
from analogai.foundation.common.dtos.notion_expression import NotionExpression
from analogai.foundation.common.logging.app_logger import app_logger
from analogai.foundation.common.utils.statement_serializers import serialize_statement
from typeguard import typechecked


class BeliefService:
    def __init__(
        self,
        statement_service: StatementServiceBase,
        notion_service: NotionService,
        categorical_deduction_service=None,
        conditional_deduction_service=None,
    ):
        if statement_service is None:
            raise ValueError("statement_service is required")
        self.statement_service = statement_service
        self.notion_service = notion_service
        self._categorical_deduction_service = categorical_deduction_service
        self._conditional_deduction_service = conditional_deduction_service
        self._belief_assembler = BeliefAssembler()
        self._belief_filterer = BeliefFilterer(
            statement_service=statement_service,
            notion_service=notion_service,
            belief_assembler=self._belief_assembler,
        )
        self._deduction_applier = BeliefDeductionApplier(
            belief_assembler=self._belief_assembler,
            belief_key_builder=self._belief_filterer.belief_key_from_statement,
            categorical_deduction_service=categorical_deduction_service,
            conditional_deduction_service=conditional_deduction_service,
            max_deduction_depth=5,
        )

    def set_categorical_deduction_service(self, service) -> None:
        self._categorical_deduction_service = service
        self._deduction_applier.set_categorical_deduction_service(service)

    def set_conditional_deduction_service(self, service) -> None:
        self._conditional_deduction_service = service
        self._deduction_applier.set_conditional_deduction_service(service)

    def _base_beliefs_for_agent(
        self,
        agent_id: str,
        modifier: Optional[Modifier],
        quantity: Optional[float],
    ) -> Dict[BeliefKey, Belief]:
        return self._belief_filterer.base_beliefs_for_agent(agent_id, modifier, quantity)

    def _belief_key_from_values(
        self,
        agent_id: str,
        subject_id: str,
        object_id: str,
        relation: Relation,
        modifier: Optional[Modifier],
    ) -> BeliefKey:
        return self._belief_filterer.belief_key_from_values(
            agent_id=agent_id,
            subject_id=subject_id,
            object_id=object_id,
            relation=relation,
            modifier=modifier,
        )

    def _belief_key_from_statement(self, statement: DirectStatement) -> BeliefKey:
        return self._belief_filterer.belief_key_from_statement(statement)

    def _beliefs_for_agent(
        self,
        agent_id: str,
        modifier: Optional[Modifier],
        quantity: Optional[float],
        user_id: Optional[str] = None,
    ) -> Dict[BeliefKey, Belief]:
        beliefs_by_key = self._belief_filterer.base_beliefs_for_agent(agent_id, modifier, quantity)
        return self._deduction_applier.apply_deductions(
            beliefs_by_key,
            agent_id=agent_id,
            modifier=modifier,
            quantity=quantity,
            user_id=user_id,
        )

    def get_base_beliefs_for_agent(
        self,
        agent_id: str,
        modifier: Optional[Modifier] = None,
        quantity: Optional[float] = None,
    ) -> List[Belief]:
        beliefs_by_key = self._belief_filterer.base_beliefs_for_agent(agent_id, modifier, quantity)
        return list(beliefs_by_key.values())


    @typechecked
    def add_statement_to_belief(self, statement: DirectStatement) -> Belief:
        statements = [statement]
        return self._belief_assembler.build_belief(statements)

    @typechecked
    def get_recently_updated(self, agent_id: str, since: Optional[datetime] = None) -> List[Belief]:
        if since is None:
            since = datetime.now() - timedelta(seconds=20)
        beliefs = self.get_all_for_agent(agent_id)
        return [belief for belief in beliefs if belief.updated_at >= since or belief.created_at >= since]

    @typechecked
    def find(self, subject_notion: Notion, complement_notion: Notion,
             relation: Relation,
             modifier: Optional[Modifier] = None,
             quantity: Optional[float] = None,
             user_id: Optional[str] = None) -> Optional[Belief]:
        modifier = self._belief_filterer.apply_quantity_modifier(modifier, quantity)
        belief_key = self._belief_filterer.belief_key_from_values(
            agent_id=subject_notion.agent_id,
            subject_id=subject_notion.id,
            object_id=complement_notion.id,
            relation=relation,
            modifier=modifier,
        )
        beliefs_by_key = self._beliefs_for_agent(
            agent_id=subject_notion.agent_id,
            modifier=modifier,
            quantity=quantity,
            user_id=user_id,
        )
        belief = beliefs_by_key.get(belief_key)
        if belief is None:
            return None
        serialized = serialize_statement(
            subject_notion_caption=belief.subject_notion.caption,
            complement_notion_caption=belief.complement_notion.caption,
            relation=belief.relation,
            probability=float(belief.probability) if belief.probability is not None else 1.0,
            modifier=belief.modifier,
            subject_attributes=belief.subject_notion.attributes,
            complement_attributes=belief.complement_notion.attributes,
        )
        app_logger.debug(f"Found Belief: {serialized} (belief_id: {belief.id})")
        return belief

    @typechecked
    def find_by_expressions(self, agent_id: str, subject_notion_expression: NotionExpression,
                           complement_notion_expression: NotionExpression,
                           relation: Relation,
                           modifier: Optional[Modifier] = None,
                           user_id: Optional[str] = None) -> Optional[Belief]:
        subject_notion = self.notion_service.find(agent_id, subject_notion_expression)
        if not subject_notion:
            app_logger.warning(f"Source notion could not be found using expression '{subject_notion_expression.caption}'")
            return None
        complement_notion = self.notion_service.find(agent_id, complement_notion_expression)
        if not complement_notion:
            app_logger.warning(f"Target notion could not be found using expression '{complement_notion_expression.caption}'")
            return None
        return self.find(
            subject_notion=subject_notion,
            complement_notion=complement_notion,
            relation=relation,
            modifier=modifier,
            user_id=user_id,
        )

    @typechecked
    def search_by_notions(self, subject_notion: Notion, complement_notion: Notion,
                          modifier: Optional[Modifier] = None) -> List[Belief]:
        return self._belief_filterer.search_by_notions(subject_notion, complement_notion, modifier)

    @typechecked
    def search_by_subject_notion(self, subject_notion: Notion,
                                modifier: Optional[Modifier] = None) -> List[Belief]:
        return self._belief_filterer.search_by_subject_notion(subject_notion, modifier)

    @typechecked
    def search_by_complement_notion(self, complement_notion: Notion,
                                modifier: Optional[Modifier] = None) -> List[Belief]:
        return self._belief_filterer.search_by_complement_notion(complement_notion, modifier)

    @typechecked
    def search_by_subject_notion_and_relation(self, subject_notion: Notion,
                                                       relation: Relation,
                                                       modifier: Optional[Modifier] = None) -> List[Belief]:
        return self._belief_filterer.search_by_subject_notion_and_relation(subject_notion, relation, modifier)

    @typechecked
    def search_by_complement_notion_and_relation(self, complement_notion: Notion,
                                                      relation: Relation,
                                                      modifier: Optional[Modifier] = None) -> List[Belief]:
        return self._belief_filterer.search_by_complement_notion_and_relation(complement_notion, relation, modifier)

    @typechecked
    def search_by_subject_notion_expression(self, agent_id: str, subject_notion_expression: NotionExpression,
                                            modifier: Optional[Modifier] = None) -> Optional[List[Belief]]:
        return self._belief_filterer.search_by_subject_notion_expression(agent_id, subject_notion_expression, modifier)

    @typechecked
    def search_by_complement_notion_expression(self, agent_id: str, complement_notion_expression: NotionExpression,
                                          modifier: Optional[Modifier] = None) -> Optional[List[Belief]]:
        return self._belief_filterer.search_by_complement_notion_expression(agent_id, complement_notion_expression, modifier)

    @typechecked
    def forget(self, belief: Belief) -> None:
        serialized = serialize_statement(
            subject_notion_caption=belief.subject_notion.caption,
            complement_notion_caption=belief.complement_notion.caption,
            relation=belief.relation,
            probability=float(belief.probability) if belief.probability is not None else 1.0,
            modifier=belief.modifier,
            subject_attributes=belief.subject_notion.attributes,
            complement_attributes=belief.complement_notion.attributes,
        )
        app_logger.debug(f"Skipped persistence for Belief: {serialized} (belief_id: {belief.id})")

    @typechecked
    def get_all_for_agent(
        self,
        agent_id: str,
        user_id: Optional[str] = None,
    ) -> List[Belief]:
        beliefs_by_key = self._beliefs_for_agent(
            agent_id=agent_id,
            modifier=None,
            quantity=None,
            user_id=user_id,
        )
        beliefs = list(beliefs_by_key.values())
        app_logger.debug(f"Found {len(beliefs)} beliefs")
        return beliefs


    @typechecked
    def search_by_criteria(self, agent_id: str, clause: Dict["Criterion", Any]) -> bool:
        from analogai.foundation.common.utils.belief_criteria_matcher import BeliefCriteriaMatcher
        beliefs = self.get_all_for_agent(agent_id)
        return any(BeliefCriteriaMatcher.matches_clause(belief, clause) for belief in beliefs)

    @typechecked
    def search_by_modifier(self, modifier: Modifier) -> List[Belief]:
        app_logger.warning("search_by_modifier is not supported without agent scope in live mode")
        return []


import uuid
from typing import List, Optional

from analogai.foundation.core.value_objects.cause import Cause
from analogai.foundation.core.value_objects.effect import Effect
from analogai.foundation.common.enums.criterion import Criterion
from analogai.foundation.common.enums.certainty import certainty_from_value


class HypotheticalStatement:
    def __init__(
        self,
        agent_id: str,
        causes: List[Cause],
        effect: Effect,
        probability: float = 1.0,
        validity: float = 1.0,
        effect_belief_id: Optional[str] = None,
        id: Optional[str] = None,
    ):
        self.id = id if id else str(uuid.uuid4())
        self.agent_id = agent_id
        self.causes = list(causes or [])
        self.effect = effect
        certainty_value = None
        if effect is not None and getattr(effect, "criteria", None):
            certainty_value = effect.criteria.get(Criterion.Certainty)
        self.probability = certainty_from_value(certainty_value).to_probability()
        self.validity = validity
        self.effect_belief_id = effect_belief_id


from analogai.foundation.modules.agent.agent_service import AgentService
from .repository_selector import create_repository_strategy


class AgentServiceFactory:
    def __init__(self, repository=None, environment: str | None = None):
        self.repository = repository or create_repository_strategy(environment).get_agent_repository()
        self.service = AgentService(self.repository)

    def get_service(self) -> AgentService:
        return self.service

    def get_repository(self):
        return self.repository



from .belief_expression_serializer import serialize_belief_expression
from .categorical_deduction_expression_serializer import serialize_categorical_deduction_expression
from .hypothetical_statement_expression_serializer import (
    serialize_hypothetical_statement_expression,
)

__all__ = [
    "serialize_belief_expression",
    "serialize_categorical_deduction_expression",
    "serialize_hypothetical_statement_expression",
]



from analogai.foundation.modules.conditionals.hypothetical_statement_service import HypotheticalStatementService
from .repository_selector import create_repository_strategy



class HypotheticalStatementServiceFactory:
    def __init__(
        self,
        repository=None,
        environment: str | None = None,
    ):
        self.repository = repository or create_repository_strategy(environment).get_hypothetical_statement_repository()
        self.service = HypotheticalStatementService(
            hypothetical_statement_repository=self.repository,
        )

    def get_service(self) -> HypotheticalStatementService:
        return self.service

    def get_repository(self):
        return self.repository


# from datetime import datetime, timedelta
# from typing import Optional, Tuple
# import re
# from analogai.foundation.core.value_objects.time import Time


# class TimeParser:
    
#     _NUMBER_WORDS = {
#         'zero': 0, 'one': 1, 'two': 2, 'three': 3, 'four': 4, 'five': 5,
#         'six': 6, 'seven': 7, 'eight': 8, 'nine': 9, 'ten': 10,
#         'eleven': 11, 'twelve': 12, 'thirteen': 13, 'fourteen': 14,
#         'fifteen': 15, 'sixteen': 16, 'seventeen': 17, 'eighteen': 18,
#         'nineteen': 19, 'twenty': 20, 'thirty': 30, 'forty': 40, 'fifty': 50,
#         'sixty': 60, 'seventy': 70, 'eighty': 80, 'ninety': 90
#     }
    
#     @staticmethod
#     def _convert_number_word_to_int(word: str) -> Optional[int]:
#         word = word.strip().lower()
        
#         if word in ['a', 'an']:
#             return 1
        
#         if word in TimeParser._NUMBER_WORDS:
#             return TimeParser._NUMBER_WORDS[word]
        
#         if '-' in word:
#             parts = word.split('-')
#             if len(parts) == 2:
#                 first = TimeParser._NUMBER_WORDS.get(parts[0])
#                 second = TimeParser._NUMBER_WORDS.get(parts[1])
#                 if first is not None and second is not None:
#                     if first >= 20 and second < 20:
#                         return first + second
        
#         return None
    
#     _TIME_UNITS = {
#         'second': 1, 'seconds': 1, 'sec': 1, 'secs': 1,
#         'minute': 60, 'minutes': 60, 'min': 60, 'mins': 60,
#         'hour': 3600, 'hours': 3600, 'hr': 3600, 'hrs': 3600,
#         'day': 86400, 'days': 86400,
#         'week': 604800, 'weeks': 604800,
#         'month': 2592000, 'months': 2592000,
#         'year': 31536000, 'years': 31536000
#     }
    
#     @staticmethod
#     def parse(time_str: str, timezone: Optional[str] = None) -> Optional[Time]:
#         if not time_str:
#             return None
        
#         time_str = time_str.strip().lower()
        
#         from analogai.foundation.common.utils.time_utils.timezone_utils import TimezoneUtils
        
#         if time_str == "now":
#             now = TimezoneUtils.get_timezone_aware_datetime(timezone)
#             return Time(
#                 year=int(now.year),
#                 month=int(now.month),
#                 day=int(now.day),
#                 hour=int(now.hour),
#                 minute=None
#             )
        
#         now = TimezoneUtils.get_timezone_aware_datetime(timezone)
        
#         relative_time = TimeParser._parse_relative_time(time_str, now)
#         if relative_time:
#             return relative_time
        
#         if time_str == "today":
#             return Time(year=int(now.year), month=int(now.month), day=int(now.day))
#         elif time_str == "tomorrow":
#             tomorrow = now + timedelta(days=1)
#             return Time(year=int(tomorrow.year), month=int(tomorrow.month), day=int(tomorrow.day))
#         elif time_str == "yesterday":
#             yesterday = now - timedelta(days=1)
#             return Time(year=int(yesterday.year), month=int(yesterday.month), day=int(yesterday.day))
        
#         try:
#             if " " in time_str:
#                 parts = time_str.split(" ", 1)
#                 date_part = parts[0]
#                 time_part = parts[1] if len(parts) > 1 else None
                
#                 if date_part in ["today", "tomorrow", "yesterday"]:
#                     base_date = now
#                     if date_part == "tomorrow":
#                         base_date = now + timedelta(days=1)
#                     elif date_part == "yesterday":
#                         base_date = now - timedelta(days=1)
                    
#                     if time_part and ":" in time_part:
#                         time_parts = time_part.split(":")
#                         hour = int(time_parts[0])
#                         minute = int(time_parts[1]) if len(time_parts) > 1 else 0
#                         return Time(
#                             year=int(base_date.year),
#                             month=int(base_date.month),
#                             day=int(base_date.day),
#                             hour=hour,
#                             minute=minute
#                         )
#                     else:
#                         return Time(
#                             year=int(base_date.year),
#                             month=int(base_date.month),
#                             day=int(base_date.day)
#                         )
#             elif ":" in time_str:
#                 time_parts = time_str.split(":")
#                 hour = int(time_parts[0])
#                 minute = int(time_parts[1]) if len(time_parts) > 1 else 0
#                 return Time(
#                     year=int(now.year),
#                     month=int(now.month),
#                     day=int(now.day),
#                     hour=hour,
#                     minute=minute
#                 )
#             elif "-" in time_str:
#                 parts = time_str.split("-")
#                 if len(parts) == 3:
#                     year = int(parts[0])
#                     month = int(parts[1])
#                     day = int(parts[2])
#                     return Time(year=year, month=month, day=day)
#         except (ValueError, AttributeError, IndexError):
#             pass
        
#         return None
    
#     @staticmethod
#     def _parse_relative_time(time_str: str, now: datetime) -> Optional[Time]:
#         time_str = time_str.strip().lower()
        
#         is_past = 'ago' in time_str
#         has_in_prefix = time_str.startswith('in ')
        
#         if has_in_prefix:
#             time_str = time_str[3:].strip()
#             is_past = False
        
#         pattern = r'\b(a|an|one|\d+|\w+(?:-\w+)?)\s+(second|seconds|sec|secs|minute|minutes|min|mins|hour|hours|hr|hrs|day|days|week|weeks|month|months|year|years)\b'
#         match = re.search(pattern, time_str)
        
#         if not match:
#             return None
        
#         quantity_str = match.group(1).strip().lower()
#         unit_str = match.group(2).strip().lower()
        
#         if unit_str not in TimeParser._TIME_UNITS:
#             return None
        
#         unit_seconds = TimeParser._TIME_UNITS[unit_str]
        
#         quantity = TimeParser._convert_number_word_to_int(quantity_str)
#         if quantity is None:
#             if quantity_str.isdigit():
#                 quantity = int(quantity_str)
#             else:
#                 return None
        
#         if quantity == 0:
#             return None
        
#         total_seconds = quantity * unit_seconds
        
#         if is_past:
#             object_time = now - timedelta(seconds=total_seconds)
#         else:
#             object_time = now + timedelta(seconds=total_seconds)
        
#         return Time(
#             year=int(object_time.year),
#             month=int(object_time.month),
#             day=int(object_time.day),
#             hour=int(object_time.hour),
#             minute=int(object_time.minute)
#         )
    
#     @staticmethod
#     def parse_day_range(time_str: str, timezone: Optional[str] = None) -> Tuple[Optional[Time], Optional[Time]]:
#         from analogai.foundation.common.utils.time_utils.timezone_utils import TimezoneUtils
        
#         if not time_str:
#             return None, None
        
#         time_str = time_str.strip().lower()
#         now = TimezoneUtils.get_timezone_aware_datetime(timezone)
        
#         if time_str == "now":
#             day_time = Time(
#                 year=int(now.year),
#                 month=int(now.month),
#                 day=int(now.day),
#                 hour=int(now.hour),
#                 minute=None
#             )
#             return day_time, day_time
#         elif time_str == "today":
#             start_time = Time(year=int(now.year), month=int(now.month), day=int(now.day), hour=0, minute=0)
#             end_time = Time(year=int(now.year), month=int(now.month), day=int(now.day), hour=23, minute=59)
#             return start_time, end_time
#         elif time_str == "tomorrow":
#             tomorrow = now + timedelta(days=1)
#             start_time = Time(year=int(tomorrow.year), month=int(tomorrow.month), day=int(tomorrow.day), hour=0, minute=0)
#             end_time = Time(year=int(tomorrow.year), month=int(tomorrow.month), day=int(tomorrow.day), hour=23, minute=59)
#             return start_time, end_time
#         elif time_str == "yesterday":
#             yesterday = now - timedelta(days=1)
#             start_time = Time(year=int(yesterday.year), month=int(yesterday.month), day=int(yesterday.day), hour=0, minute=0)
#             end_time = Time(year=int(yesterday.year), month=int(yesterday.month), day=int(yesterday.day), hour=23, minute=59)
#             return start_time, end_time
        
#         parsed_time = TimeParser.parse(time_str, timezone)
#         return parsed_time, None



from __future__ import annotations

from typing import List

from analogai.foundation.common.logging.app_logger import app_logger
from analogai.foundation.infrastructure.embeddings.http_client import post_json
from analogai.foundation.settings import config


class OpenAIEmbeddingProvider:
    def __init__(self, api_key: str | None = None, model_name: str | None = None):
        resolved_api_key = api_key or (config.get("OPENAI_API_KEY", "") or "")
        resolved_model_name = model_name or (config.get("embed.model", "") or "")
        if not resolved_api_key:
            raise ValueError("OPENAI_API_KEY is required for OpenAI embeddings")
        if not resolved_model_name:
            raise ValueError("EMBEDDING_MODEL_NAME is required for OpenAI embeddings")
        self._api_key = resolved_api_key
        self._model_name = resolved_model_name
        self._dimension: int | None = None

    @property
    def dimension(self) -> int | None:
        return self._dimension

    def encode(self, text: str) -> List[float]:
        if not text or not text.strip():
            if self._dimension is None:
                return []
            return [0.0] * self._dimension
        payload = {"model": self._model_name, "input": text}
        response = post_json(
            "https://api.openai.com/v1/embeddings",
            payload,
            headers={
                "Authorization": f"Bearer {self._api_key}",
                "Content-Type": "application/json",
            },
        )
        embedding = (response.get("data") or [{}])[0].get("embedding") or []
        if embedding and self._dimension is None:
            self._dimension = len(embedding)
        if not embedding:
            app_logger.warning("OpenAI embeddings returned empty vector")
        return embedding

    def encode_batch(self, texts: List[str]) -> List[List[float]]:
        if not texts:
            return []
        payload = {"model": self._model_name, "input": texts}
        response = post_json(
            "https://api.openai.com/v1/embeddings",
            payload,
            headers={
                "Authorization": f"Bearer {self._api_key}",
                "Content-Type": "application/json",
            },
        )
        data = response.get("data") or []
        embeddings = [item.get("embedding") or [] for item in data]
        if embeddings and self._dimension is None and embeddings[0]:
            self._dimension = len(embeddings[0])
        return embeddings



from analogai.foundation.extensions.conversation.services.conversation_service import ConversationService
from analogai.foundation.extensions.conversation.repositories import InMemoryConversationRepositoryImpl, MongoDBConversationRepositoryImpl
from analogai.foundation.common.enums.environment import Environment
from analogai.foundation.settings import config

class ConversationServiceFactory:
    def create(self) -> ConversationService:
        env = Environment(config.get("ENVIRONMENT"))
        if env == Environment.PRODUCTION:
            repository = MongoDBConversationRepositoryImpl()
        elif env in {Environment.DEVELOPMENT, Environment.TESTING}:
            repository = InMemoryConversationRepositoryImpl()
        else:
            raise ValueError(f"No repository implementation for environment: {env}")
        return ConversationService(repository)



from typing import Dict
from datetime import datetime
from analogai.foundation.extensions.fallback import Fallback


class FallbackStorageMapper:
    @staticmethod
    def to_dict(fallback: Fallback) -> Dict:
        return {
            "id": fallback.id,
            "user_id": fallback.user_id,
            "agent_id": fallback.agent_id,
            "user_question": fallback.user_question,
            "timestamp": fallback.timestamp.isoformat() if fallback.timestamp else None
        }

    @staticmethod
    def from_dict(data: Dict) -> Fallback:
        timestamp = None
        if data.get("timestamp"):
            if isinstance(data["timestamp"], str):
                timestamp = datetime.fromisoformat(data["timestamp"])
            elif isinstance(data["timestamp"], datetime):
                timestamp = data["timestamp"]

        return Fallback(
            id=data.get("id"),
            user_id=data["user_id"],
            agent_id=data["agent_id"],
            user_question=data["user_question"],
            timestamp=timestamp
        )



from datetime import datetime
from typing import Dict

from analogai.foundation.modules.direct_statement.direct_statement import DirectStatement
from analogai.foundation.core.value_objects.relation import Relation
from analogai.foundation.common.utils.modifier_utils import ModifierBuilder
from analogai.foundation.common.enums.deontic_modality import DeonticModality
from analogai.foundation.infrastructure.mappers.notion_storage_mapper import NotionStorageMapper
from analogai.foundation.infrastructure.mappers.time_storage_mapper import TimeStorageMapper


class StatementStorageMapper:
    @staticmethod
    def to_dict(statement: DirectStatement) -> Dict:
        result = {
            "id": statement.id,
            "user_id": statement.user_id,
            "agent_id": statement.agent_id,
            "created_at": statement.created_at.isoformat() if statement.created_at else None,
            "subject_notion": NotionStorageMapper.to_dict(statement.subject_notion),
            "relation": statement.relation.value,
            "complement_notion": NotionStorageMapper.to_dict(statement.complement_notion),
            "probability": statement.probability,
            "validity": statement.validity,
        }
        if statement.quantity is not None:
            result["quantity"] = statement.quantity
        if statement.modifier:
            modifier = statement.modifier
            if modifier.location:
                result["location"] = modifier.location
            if modifier.from_time:
                result.update(TimeStorageMapper.flatten_to_dict(modifier.from_time, "from_time"))
            if modifier.to_time:
                result.update(TimeStorageMapper.flatten_to_dict(modifier.to_time, "to_time"))
            if modifier.deontic_modality:
                result["deontic_modality"] = modifier.deontic_modality.value
            if modifier.attributes is not None:
                result["attributes"] = [
                    {"tag": attribute.tag, "value": attribute.value, "unit": attribute.unit}
                    for attribute in modifier.attributes
                ]
        return result

    @staticmethod
    def from_dict(data: Dict) -> DirectStatement:
        subject_notion = NotionStorageMapper.from_dict(data.get("subject_notion", {}))
        relation_str = data.get("relation") or data.get("relationship_type") or data.get("relationship")
        relation = Relation.from_string(relation_str)
        complement_notion = NotionStorageMapper.from_dict(data.get("complement_notion", {}))
        modifier = None
        from_time = TimeStorageMapper.unflatten_from_dict(data, "from_time") or (
            data.get("from_time") and TimeStorageMapper.from_dict(data["from_time"])
        )
        to_time = TimeStorageMapper.unflatten_from_dict(data, "to_time") or (
            data.get("to_time") and TimeStorageMapper.from_dict(data["to_time"])
        )
        location = data.get("location")
        deontic_modality = data.get("deontic_modality")
        if location or from_time or to_time or deontic_modality:
            builder = ModifierBuilder()
            if location:
                builder.with_location(location)
            if from_time:
                builder.with_from_time(
                    year=from_time.year,
                    month=from_time.month,
                    day=from_time.day,
                    hour=from_time.hour,
                    minute=from_time.minute,
                )
            if to_time:
                builder.with_to_time(
                    year=to_time.year,
                    month=to_time.month,
                    day=to_time.day,
                    hour=to_time.hour,
                    minute=to_time.minute,
                )
            if deontic_modality:
                builder.with_deontic_modality(DeonticModality(deontic_modality))
            modifier = builder.build()
        attributes_payload = data.get("attributes")
        if attributes_payload is None:
            attributes_payload = data.get("measures")
        if attributes_payload is not None:
            from analogai.foundation.core.value_objects.attribute import Attribute
            attributes = [
                Attribute(tag=m.get("tag") or m.get("kind"), value=m.get("value"), unit=m.get("unit"))
                for m in attributes_payload or []
            ]
            if modifier is None:
                modifier = ModifierBuilder().build()
            modifier.attributes = attributes
        if data.get("quantity") is not None:
            if modifier is None:
                modifier = ModifierBuilder().build()
            modifier.quantity = data.get("quantity")
        created_at = None
        if data.get("created_at"):
            raw = data["created_at"]
            try:
                created_at = datetime.fromisoformat(raw)
            except TypeError:
                created_at = raw
        return DirectStatement(
            id=data.get("id"),
            user_id=data["user_id"],
            agent_id=data["agent_id"],
            subject_notion=subject_notion,
            relation=relation,
            complement_notion=complement_notion,
            modifier=modifier,
            probability=data.get("probability", 1.0),
            validity=data.get("validity", 1.0),
            created_at=created_at,
        )


from abc import ABC, abstractmethod
from typing import Optional
from analogai.foundation.modules.user.user import User

class UserRepository(ABC):
    @abstractmethod
    def create(self, user: User) -> User:
        pass

    @abstractmethod
    def find_by_id(self, id: str) -> Optional[User]:
        pass

    @abstractmethod
    def update(self, user: User) -> User:
        pass

    @abstractmethod
    def find_by_email(self, email: str) -> Optional[User]:
        pass

    @abstractmethod
    def delete(self, id: str) -> bool:
        pass


from analogai.foundation.modules.agent.agent_repository import AgentRepository
from analogai.foundation.modules.agent.agent import Agent
from analogai.foundation.modules.user.user import User
from typeguard import typechecked
from typing import Optional, List

class AgentService:
    def __init__(self, agent_repository: AgentRepository):
        self.agent_repository = agent_repository

    @typechecked
    def find(self, id: str) -> Optional[Agent]:
        return self.agent_repository.find_by_id(id)

    @typechecked
    def get_agents_by_creator(self, creator_id: str) -> List[Agent]:
        return self.agent_repository.find_by_creator_id(creator_id)

    @typechecked
    def create(self, creator: User, id: str = None, name: str = "", knowledge_base: str = "", roles=None) -> Agent:
        agent_params = {"creator": creator, "name": name, "knowledge_base": knowledge_base, "roles": roles or []}
        if id is not None:
            agent_params["id"] = id
        agent = Agent(**agent_params)
        return self.agent_repository.create(agent)

    @typechecked
    def update(self, agent: Agent) -> Agent:
        return self.agent_repository.update(agent)

    @typechecked
    def delete(self, id: str) -> bool:
        return self.agent_repository.delete(id)


import importlib
import os
import unittest
from unittest.mock import patch

from analogai.foundation.common.design_patterns.singleton import Singleton
from analogai.foundation.common.utils.env_utils import load_envs_in_order
from analogai.foundation.settings import config


class TestEmbeddingServiceFunctionality(unittest.TestCase):
    def setUp(self):
        self._env_backup = dict(os.environ)
        load_envs_in_order()

    def tearDown(self):
        os.environ.clear()
        os.environ.update(self._env_backup)
        from analogai.foundation.infrastructure.services import embedding_service

        Singleton._instances.pop(embedding_service.EmbeddingService, None)

    def _reload_embedding_modules(self):
        from analogai.foundation.infrastructure.embeddings import provider_factory
        from analogai.foundation.infrastructure.services import embedding_service
        from analogai.foundation import settings

        importlib.reload(settings)
        importlib.reload(provider_factory)
        importlib.reload(embedding_service)
        return settings.config, provider_factory, embedding_service

    def test_embedding_provider_encode(self):
        _config, _provider_factory, embedding_service = self._reload_embedding_modules()
        Singleton._instances.pop(embedding_service.EmbeddingService, None)
        service = embedding_service.EmbeddingService()
        vector = service.encode("hello")
        self.assertTrue(len(vector) > 0)

    def test_embedding_provider_batch_encode(self):
        _config, _provider_factory, embedding_service = self._reload_embedding_modules()
        environment = _config.get("ENVIRONMENT")
        use_vector_search = environment not in {"test", "testing"} and bool(
            _config.get("vector.enabled", True)
        )
        if not use_vector_search:
            self.skipTest("USE_VECTOR_SEARCH is false; skipping embedding tests")

        Singleton._instances.pop(embedding_service.EmbeddingService, None)
        service = embedding_service.EmbeddingService()
        batch = service.encode_batch(["a", "b"])
        self.assertEqual(len(batch), 2)
        self.assertTrue(len(batch[0]) > 0)


if __name__ == "__main__":
    unittest.main()
